# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Workspace module."""

import os
import subprocess
import sys
from pathlib import Path
from typing import Tuple

from click import Choice
from click import confirm
from click import echo
from click import group
from click import option
from click import pass_context
from click import Path as ClickPath

from openfl.utilities.path_check import is_directory_traversal
from openfl.utilities.workspace import dump_requirements_file


@group()
@pass_context
def workspace(context):
    """Manage Federated Learning Workspaces."""
    context.obj['group'] = 'workspace'


def create_dirs(prefix):
    """Create workspace directories."""
    from shutil import copyfile

    from openfl.interface.cli_helper import WORKSPACE

    echo('Creating Workspace Directories')

    (prefix / 'cert').mkdir(parents=True, exist_ok=True)  # certifications
    (prefix / 'data').mkdir(parents=True, exist_ok=True)  # training data
    (prefix / 'logs').mkdir(parents=True, exist_ok=True)  # training logs
    (prefix / 'save').mkdir(parents=True, exist_ok=True)  # model weight saves / initialization
    (prefix / 'src').mkdir(parents=True, exist_ok=True)  # model code

    copyfile(WORKSPACE / 'workspace' / '.workspace', prefix / '.workspace')


def create_temp(prefix, template):
    """Create workspace templates."""
    from shutil import ignore_patterns

    from openfl.interface.cli_helper import copytree
    from openfl.interface.cli_helper import WORKSPACE

    echo('Creating Workspace Templates')

    copytree(src=WORKSPACE / template, dst=prefix, dirs_exist_ok=True,
             ignore=ignore_patterns('__pycache__'))  # from template workspace
    apply_template_plan(prefix, template)


def get_templates():
    """Grab the default templates from the distribution."""
    from openfl.interface.cli_helper import WORKSPACE

    return [d.name for d in WORKSPACE.glob('*') if d.is_dir()
            and d.name not in ['__pycache__', 'workspace']]


@workspace.command(name='create')
@option('--prefix', required=True,
        help='Workspace name or path', type=ClickPath())
@option('--template', required=True, type=Choice(get_templates()))
def create_(prefix, template):
    """Create the workspace."""
    if is_directory_traversal(prefix):
        echo('Workspace name or path is out of the openfl workspace scope.')
        sys.exit(1)
    create(prefix, template)


def create(prefix, template):
    """Create federated learning workspace."""
    from os.path import isfile
    from subprocess import check_call
    from sys import executable

    from openfl.interface.cli_helper import print_tree
    from openfl.interface.cli_helper import OPENFL_USERDIR

    if not OPENFL_USERDIR.exists():
        OPENFL_USERDIR.mkdir()

    prefix = Path(prefix).absolute()

    create_dirs(prefix)
    create_temp(prefix, template)

    requirements_filename = 'requirements.txt'

    if isfile(f'{str(prefix)}/{requirements_filename}'):
        check_call([
            executable, '-m', 'pip', 'install', '-r',
            f'{prefix}/requirements.txt'], shell=False)
        echo(f'Successfully installed packages from {prefix}/requirements.txt.')
    else:
        echo('No additional requirements for workspace defined. Skipping...')
    prefix_hash = _get_dir_hash(str(prefix.absolute()))
    with open(OPENFL_USERDIR / f'requirements.{prefix_hash}.txt', 'w') as f:
        check_call([executable, '-m', 'pip', 'freeze'], shell=False, stdout=f)

    print_tree(prefix, level=3)


@workspace.command(name='export')
@option('-o', '--pip-install-options', required=False,
        type=str, multiple=True, default=tuple,
        help='Options for remote pip install. '
             'You may pass several options in quotation marks alongside with arguments, '
             'e.g. -o "--find-links source.site"')
def export_(pip_install_options: Tuple[str]):
    """Export federated learning workspace."""
    from os import getcwd
    from os import makedirs
    from os.path import basename
    from os.path import join
    from shutil import copy2
    from shutil import copytree
    from shutil import ignore_patterns
    from shutil import make_archive
    from tempfile import mkdtemp

    from plan import freeze_plan
    from openfl.interface.cli_helper import WORKSPACE

    plan_file = Path('plan/plan.yaml').absolute()
    try:
        freeze_plan(plan_file)
    except Exception:
        echo(f'Plan file "{plan_file}" not found. No freeze performed.')

    # Dump requirements.txt
    dump_requirements_file(prefixes=pip_install_options, keep_original_prefixes=True)

    archive_type = 'zip'
    archive_name = basename(getcwd())
    archive_file_name = archive_name + '.' + archive_type

    # Aggregator workspace
    tmp_dir = join(mkdtemp(), 'openfl', archive_name)

    ignore = ignore_patterns(
        '__pycache__', '*.crt', '*.key', '*.csr', '*.srl', '*.pem', '*.pbuf')

    # We only export the minimum required files to set up a collaborator
    makedirs(f'{tmp_dir}/save', exist_ok=True)
    makedirs(f'{tmp_dir}/logs', exist_ok=True)
    makedirs(f'{tmp_dir}/data', exist_ok=True)
    copytree('./src', f'{tmp_dir}/src', ignore=ignore)  # code
    copytree('./plan', f'{tmp_dir}/plan', ignore=ignore)  # plan
    copy2('./requirements.txt', f'{tmp_dir}/requirements.txt')  # requirements

    try:
        copy2('.workspace', tmp_dir)  # .workspace
    except FileNotFoundError:
        echo('\'.workspace\' file not found.')
        if confirm('Create a default \'.workspace\' file?'):
            copy2(WORKSPACE / 'workspace' / '.workspace', tmp_dir)
        else:
            echo('To proceed, you must have a \'.workspace\' '
                 'file in the current directory.')
            raise

    # Create Zip archive of directory
    echo('\n 🗜️ Preparing workspace distribution zip file')
    make_archive(archive_name, archive_type, tmp_dir)

    echo(f'\n ✔️ Workspace exported to archive: {archive_file_name}')


@workspace.command(name='import')
@option('--archive', required=True,
        help='Zip file containing workspace to import',
        type=ClickPath(exists=True))
def import_(archive):
    """Import federated learning workspace."""
    from os import chdir
    from os.path import basename
    from os.path import isfile
    from shutil import unpack_archive
    from subprocess import check_call
    from sys import executable

    archive = Path(archive).absolute()

    dir_path = basename(archive).split('.')[0]
    unpack_archive(archive, extract_dir=dir_path)
    chdir(dir_path)

    requirements_filename = 'requirements.txt'

    if isfile(requirements_filename):
        check_call([
            executable, '-m', 'pip', 'install', '--upgrade', 'pip'],
            shell=False)
        check_call([
            executable, '-m', 'pip', 'install', '-r', 'requirements.txt'],
            shell=False)
    else:
        echo('No ' + requirements_filename + ' file found.')

    echo(f'Workspace {archive} has been imported.')
    echo('You may need to copy your PKI certificates to join the federation.')


@workspace.command(name='certify')
def certify_():
    """Create certificate authority for federation."""
    certify()


def certify():
    """Create certificate authority for federation."""
    from cryptography.hazmat.primitives import serialization

    from openfl.cryptography.ca import generate_root_cert
    from openfl.cryptography.ca import generate_signing_csr
    from openfl.cryptography.ca import sign_certificate
    from openfl.interface.cli_helper import CERT_DIR

    echo('Setting Up Certificate Authority...\n')

    echo('1.  Create Root CA')
    echo('1.1 Create Directories')

    (CERT_DIR / 'ca/root-ca/private').mkdir(
        parents=True, exist_ok=True, mode=0o700)
    (CERT_DIR / 'ca/root-ca/db').mkdir(parents=True, exist_ok=True)

    echo('1.2 Create Database')

    with open(CERT_DIR / 'ca/root-ca/db/root-ca.db', 'w') as f:
        pass  # write empty file
    with open(CERT_DIR / 'ca/root-ca/db/root-ca.db.attr', 'w') as f:
        pass  # write empty file

    with open(CERT_DIR / 'ca/root-ca/db/root-ca.crt.srl', 'w') as f:
        f.write('01')  # write file with '01'
    with open(CERT_DIR / 'ca/root-ca/db/root-ca.crl.srl', 'w') as f:
        f.write('01')  # write file with '01'

    echo('1.3 Create CA Request and Certificate')

    root_crt_path = 'ca/root-ca.crt'
    root_key_path = 'ca/root-ca/private/root-ca.key'

    root_private_key, root_cert = generate_root_cert()

    # Write root CA certificate to disk
    with open(CERT_DIR / root_crt_path, 'wb') as f:
        f.write(root_cert.public_bytes(
            encoding=serialization.Encoding.PEM,
        ))

    with open(CERT_DIR / root_key_path, 'wb') as f:
        f.write(root_private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        ))

    echo('2.  Create Signing Certificate')
    echo('2.1 Create Directories')

    (CERT_DIR / 'ca/signing-ca/private').mkdir(
        parents=True, exist_ok=True, mode=0o700)
    (CERT_DIR / 'ca/signing-ca/db').mkdir(parents=True, exist_ok=True)

    echo('2.2 Create Database')

    with open(CERT_DIR / 'ca/signing-ca/db/signing-ca.db', 'w') as f:
        pass  # write empty file
    with open(CERT_DIR / 'ca/signing-ca/db/signing-ca.db.attr', 'w') as f:
        pass  # write empty file

    with open(CERT_DIR / 'ca/signing-ca/db/signing-ca.crt.srl', 'w') as f:
        f.write('01')  # write file with '01'
    with open(CERT_DIR / 'ca/signing-ca/db/signing-ca.crl.srl', 'w') as f:
        f.write('01')  # write file with '01'

    echo('2.3 Create Signing Certificate CSR')

    signing_csr_path = 'ca/signing-ca.csr'
    signing_crt_path = 'ca/signing-ca.crt'
    signing_key_path = 'ca/signing-ca/private/signing-ca.key'

    signing_private_key, signing_csr = generate_signing_csr()

    # Write Signing CA CSR to disk
    with open(CERT_DIR / signing_csr_path, 'wb') as f:
        f.write(signing_csr.public_bytes(
            encoding=serialization.Encoding.PEM,
        ))

    with open(CERT_DIR / signing_key_path, 'wb') as f:
        f.write(signing_private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption()
        ))

    echo('2.4 Sign Signing Certificate CSR')

    signing_cert = sign_certificate(signing_csr, root_private_key, root_cert.subject, ca=True)

    with open(CERT_DIR / signing_crt_path, 'wb') as f:
        f.write(signing_cert.public_bytes(
            encoding=serialization.Encoding.PEM,
        ))

    echo('3   Create Certificate Chain')

    # create certificate chain file by combining root-ca and signing-ca
    with open(CERT_DIR / 'cert_chain.crt', 'w') as d:
        with open(CERT_DIR / 'ca/root-ca.crt') as s:
            d.write(s.read())
        with open(CERT_DIR / 'ca/signing-ca.crt') as s:
            d.write(s.read())

    echo('\nDone.')


def _get_requirements_dict(txtfile):
    with open(txtfile, 'r') as snapshot:
        snapshot_dict = {}
        for line in snapshot:
            try:
                # 'pip freeze' generates requirements with exact versions
                k, v = line.split('==')
                snapshot_dict[k] = v
            except ValueError:
                snapshot_dict[line] = None
        return snapshot_dict


def _get_dir_hash(path):
    from hashlib import sha256
    hash_ = sha256()
    hash_.update(path.encode('utf-8'))
    hash_ = hash_.hexdigest()
    return hash_


@workspace.command(name='dockerize')
@option('-b', '--base_image', required=False,
        help='The tag for openfl base image',
        default='openfl')
@option('--save/--no-save',
        required=False,
        help='Save the Docker image into the workspace',
        default=True)
@pass_context
def dockerize_(context, base_image, save):
    """
    Pack the workspace as a Docker image.

    This command is the alternative to `workspace export`.
    It should be called after plan initialization from the workspace dir.

    User is expected to be in docker group.
    If your machine is behind a proxy, make sure you set it up in ~/.docker/config.json.
    """
    import docker
    import sys
    from shutil import copyfile

    from openfl.interface.cli_helper import SITEPACKS

    # Specify the Dockerfile.workspace loaction
    openfl_docker_dir = os.path.join(SITEPACKS, 'openfl-docker')
    dockerfile_workspace = 'Dockerfile.workspace'
    # Apparently, docker's python package does not support
    # scenarios when the dockerfile is placed outside the build context
    copyfile(os.path.join(openfl_docker_dir, dockerfile_workspace), dockerfile_workspace)

    workspace_path = os.getcwd()
    workspace_name = os.path.basename(workspace_path)

    # Exporting the workspace
    context.invoke(export_)
    workspace_archive = workspace_name + '.zip'

    build_args = {
        'WORKSPACE_NAME': workspace_name,
        'BASE_IMAGE': base_image
    }

    client = docker.from_env(timeout=3600)
    echo('Building the Docker image')
    try:
        client.images.build(
            path=str(workspace_path),
            tag=workspace_name,
            buildargs=build_args,
            dockerfile=dockerfile_workspace
        )
    except docker.errors.BuildError as e:
        for log in e.build_log:
            msg = log.get('stream')
            if msg:
                echo(msg)
        echo('Failed to build the image\n' + str(e) + '\n')
        sys.exit(1)
    finally:
        os.remove(workspace_archive)
        os.remove(dockerfile_workspace)
    echo('The workspace image has been built successfully!')

    # Saving the image to a tarball
    if save:
        workspace_image_tar = workspace_name + '_image.tar'
        echo('Saving the Docker image...')
        image = client.images.get(f'{workspace_name}')
        resp = image.save(named=True)
        with open(workspace_image_tar, 'wb') as f:
            for chunk in resp:
                f.write(chunk)
        echo(f'{workspace_name} image saved to {workspace_path}/{workspace_image_tar}')


@workspace.command(name='graminize')
@option('-s', '--signing-key', required=False,
        type=lambda p: Path(p).absolute(), default='/',
        help='A 3072-bit RSA private key (PEM format) is required for signing the manifest.\n'
             'If a key is passed the gramine-sgx manifest fill be prepared.\n'
             'In option is ignored this command will build an image that can only run '
             'with gramine-direct (not in enclave).',
        )
@option('-o', '--pip-install-options', required=False,
        type=str, multiple=True, default=tuple,
        help='Options for remote pip install. '
             'You may pass several options in quotation marks alongside with arguments, '
             'e.g. -o "--find-links source.site"')
@option('--save/--no-save', required=False,
        default=True, type=bool,
        help='Dump the Docker image to an archive')
@option('--rebuild', help='Build images with `--no-cache`', is_flag=True)
@pass_context
def graminize_(context, signing_key: Path, pip_install_options: Tuple[str],
               save: bool, rebuild: bool) -> None:
    """
    Build gramine app inside a docker image.

    This command is the alternative to `workspace export`.
    It should be called after `fx plan initialize` inside the workspace dir.

    User is expected to be in docker group.
    If your machine is behind a proxy, make sure you set it up in ~/.docker/config.json.

    TODO:
    1. gramine-direct, check if a key is provided
    2. make a standalone function with `export` parametr
    """
    def open_pipe(command: str):
        echo(f'\n 📦 Executing command:\n{command}\n')
        process = subprocess.Popen(
            command,
            shell=True, stderr=subprocess.STDOUT,
            stdout=subprocess.PIPE)
        for line in process.stdout:
            echo(line)
        _ = process.communicate()  # pipe is already empty, used to get `returncode`
        if process.returncode != 0:
            raise Exception('\n ❌ Execution failed\n')

    from openfl.interface.cli_helper import SITEPACKS

    # We can build for gramine-sgx and run with gramine-direct,
    # but not vice versa.
    sgx_build = signing_key.is_file()
    if sgx_build:
        echo('\n Building SGX-ready applecation')
    else:
        echo('\n Building gramine-direct applecation')
    rebuild_option = '--no-cache' if rebuild else ''

    os.environ['DOCKER_BUILDKIT'] = '1'

    echo('\n 🐋 Building base gramine-openfl image...')
    base_dockerfile = SITEPACKS / 'openfl-gramine' / 'Dockerfile.gramine'
    base_build_command = f'docker build {rebuild_option} -t gramine_openfl -f {base_dockerfile} .'
    open_pipe(base_build_command)
    echo('\n ✔️ DONE: Building base gramine-openfl image')

    workspace_path = Path.cwd()
    workspace_name = workspace_path.name
    context.invoke(export_, pip_install_options=pip_install_options)
    workspace_archive = workspace_path / f'{workspace_name}.zip'

    grainized_ws_dockerfile = SITEPACKS / 'openfl-gramine' / 'Dockerfile.graminized.workspace'

    echo('\n 🐋 Building graminized workspace image...')
    signing_key = f'--secret id=signer-key,src={signing_key} ' if sgx_build else ''
    graminized_build_command = (
        f'docker build -t {workspace_name} {rebuild_option} '
        '--build-arg BASE_IMAGE=gramine_openfl '
        f'--build-arg WORKSPACE_ARCHIVE={workspace_archive.relative_to(workspace_path)} '
        f'--build-arg SGX_BUILD={int(sgx_build)} '
        f'{signing_key}'
        f'-f {grainized_ws_dockerfile} {workspace_path}')
    open_pipe(graminized_build_command)
    echo('\n ✔️ DONE: Building graminized workspace image')

    if save:
        echo('\n 💾 Saving the graminized workspace image...')
        save_image_command = f'docker save {workspace_name} | gzip > {workspace_name}.tar.gz'
        open_pipe(save_image_command)
        echo(f'\n ✔️ The image saved to file: {workspace_name}.tar.gz')


def apply_template_plan(prefix, template):
    """Copy plan file from template folder.

    This function unfolds default values from template plan configuration
    and writes the configuration to the current workspace.
    """
    from openfl.federated.plan import Plan
    from openfl.interface.cli_helper import WORKSPACE

    template_plan = Plan.parse(WORKSPACE / template / 'plan' / 'plan.yaml')

    Plan.dump(prefix / 'plan' / 'plan.yaml', template_plan.config)



import argparse

import numpy as np
from sklearn.exceptions import NotFittedError
from sklearn.metrics import accuracy_score

from IrisDataset import IrisDataset
from openfl.interface.interactive_api.experiment import FLExperiment, TaskInterface, ModelInterface
from openfl.interface.interactive_api.federation import Federation
from random_forest import MyRandomForestClassifier

LOG_WANDB = False

parser = argparse.ArgumentParser(description="Script")
parser.add_argument("--rounds", default=3, type=int)
parser.add_argument("--seed", default=42, type=int)
parser.add_argument("--server", default='localhost', type=str, help="server address")
args = parser.parse_args()

random_state = np.random.RandomState(args.seed)

client_id = 'api'
cert_dir = '../cert'

task_interface = TaskInterface()


@task_interface.register_fl_task(model='model', data_loader='train_loader', device='device', optimizer='optimizer')
def train(model, train_loader, device, optimizer):
    X, y = train_loader
    X, y = np.array(X), np.array(y)
    model.fit(X, y)
    pred = model.predict(X)
    metric = accuracy_score(pred, y)

    return {'accuracy': metric}


@task_interface.register_fl_task(model='model', data_loader='val_loader', device='device')
def validate(model, val_loader, device):
    try:
        if model is not None:
            X, y = val_loader
            pred = model.predict(X)
            metric = accuracy_score(pred, y)
        else:
            print("Model not found")
            metric = 0
    except NotFittedError:
        print("Model is not yet fit")
        metric = 0

    return {'accuracy': metric}


federation = Federation(client_id=client_id, director_node_fqdn=args.server, director_port='50052', tls=False)
fl_experiment = FLExperiment(federation=federation, experiment_name="RandomForest_Iris",
                             serializer_plugin='openfl.plugins.interface_serializer.dill_serializer.DillSerializer',
                             load_default_plan=False, nn=False)
model_interface = ModelInterface(
    model=MyRandomForestClassifier(),
    optimizer=None,
    framework_plugin='openfl.plugins.frameworks_adapters.generic_adapter.GenericAdapter')
federated_dataset = IrisDataset()

fl_experiment.start(
    model_provider=model_interface,
    task_keeper=task_interface,
    data_loader=federated_dataset,
    rounds_to_train=args.rounds,
    opt_treatment='CONTINUE_GLOBAL'
)

fl_experiment.stream_metrics(tensorboard_logs=False)
fl_experiment.remove_experiment_data()



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""FastEsitmatorDataLoader module."""

from .loader import DataLoader


class FastEstimatorDataLoader(DataLoader):
    """Federation Data Loader for FastEstimator."""

    def __init__(self, pipeline, **kwargs):
        """
        Instantiate the data object.

        Args:
            batch_size: Size of batches used for all data loaders
            kwargs: consumes all un-used kwargs

        Returns:
            None
        """
        self.pipeline = pipeline

        self.batch_size = pipeline.batch_size
        self.X_train = None
        self.y_train = None
        self.X_valid = None
        self.y_valid = None

        # Child classes should have init signature:
        # (self, batch_size, **kwargs), should call this __init__ and then
        # define self.X_train, self.y_train, self.X_valid, and self.y_valid

    def get_train_data_size(self):
        """
        Get total number of training samples.

        Returns:
            int: number of training samples
        """
        return len(self.pipeline.data['train'])

    def get_valid_data_size(self):
        """
        Get total number of validation samples.

        Returns:
            int: number of validation samples
        """
        return len(self.pipeline.data['test'])

    def get_feature_shape(self):
        """
        Get feature shape.

        Returns:
            tuple: shape of the input data.
        """
        return self.pipeline.data['train']['x'].shape[1:]



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Task module."""


from dataclasses import dataclass
from dataclasses import field


@dataclass
class Task:
    """Task base dataclass."""

    name: str
    function_name: str
    task_type: str
    apply_local: bool = False
    parameters: dict = field(default_factory=dict)  # We can expend it in the future


@dataclass
class TrainTask(Task):
    """TrainTask class."""

    task_type: str = 'train'


@dataclass
class ValidateTask(Task):
    """Validation Task class."""

    task_type: str = 'validate'



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Custom Assigner module."""


import logging
from collections import defaultdict

from openfl.interface.aggregation_functions import WeightedAverage

logger = logging.getLogger(__name__)


class Assigner:
    """Custom assigner class."""

    def __init__(
            self,
            *,
            assigner_function,
            aggregation_functions_by_task,
            authorized_cols,
            rounds_to_train
    ):
        """Initialize."""
        self.agg_functions_by_task = aggregation_functions_by_task
        self.agg_functions_by_task_name = {}
        self.authorized_cols = authorized_cols
        self.rounds_to_train = rounds_to_train
        self.all_tasks_for_round = defaultdict(dict)
        self.collaborators_for_task = defaultdict(lambda: defaultdict(list))
        self.collaborator_tasks = defaultdict(lambda: defaultdict(list))
        self.assigner_function = assigner_function

        self.define_task_assignments()

    def define_task_assignments(self):
        """Abstract method."""
        for round_number in range(self.rounds_to_train):
            tasks_by_collaborator = self.assigner_function(
                self.authorized_cols,
                round_number,
                number_of_callaborators=len(self.authorized_cols)
            )
            for collaborator_name, tasks in tasks_by_collaborator.items():
                self.collaborator_tasks[round_number][collaborator_name].extend(tasks)
                for task in tasks:
                    self.all_tasks_for_round[round_number][task.name] = task
                    self.collaborators_for_task[round_number][task.name].append(collaborator_name)
                    if self.agg_functions_by_task:
                        self.agg_functions_by_task_name[
                            task.name
                        ] = self.agg_functions_by_task.get(task.function_name, WeightedAverage())

    def get_tasks_for_collaborator(self, collaborator_name, round_number):
        """Abstract method."""
        return self.collaborator_tasks[round_number][collaborator_name]

    def get_collaborators_for_task(self, task_name, round_number):
        """Abstract method."""
        return self.collaborators_for_task[round_number][task_name]

    def get_all_tasks_for_round(self, round_number):
        """
        Return tasks for the current round.

        Currently all tasks are performed on each round,
        But there may be a reason to change this.
        """
        return [task.name for task in self.all_tasks_for_round[round_number].values()]

    def get_aggregation_type_for_task(self, task_name):
        """Extract aggregation type from self.tasks."""
        agg_fn = self.agg_functions_by_task_name.get(task_name, WeightedAverage())
        return agg_fn



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""TensorDB Module."""

from threading import Lock
from types import MethodType
from typing import Dict
from typing import Iterator
from typing import Optional

import numpy as np
import pandas as pd

from openfl.databases.utilities import _search, _store, _retrieve, ROUND_PLACEHOLDER
from openfl.interface.aggregation_functions import AggregationFunction
from openfl.utilities import LocalTensor
from openfl.utilities import TensorKey
from openfl.utilities import change_tags


class TensorDB:
    """
    The TensorDB stores a tensor key and the data that it corresponds to.

    It is built on top of a pandas dataframe
    for it's easy insertion, retreival and aggregation capabilities. Each
    collaborator and aggregator has its own TensorDB.
    """

    def __init__(self, nn=True, agg=True) -> None:
        """Initialize."""
        self.tensor_db = pd.DataFrame([], columns=[
            'tensor_name', 'origin', 'round', 'report', 'tags', 'nparray'
        ])
        self.tensor_db = self.tensor_db.astype({"report": bool})
        self._bind_convenience_methods()
        self.nn = nn
        self.agg = agg
        self.mutex = Lock()

    def _bind_convenience_methods(self):
        # Bind convenience methods for TensorDB dataframe to make storage, retrieval, and search easier
        if not hasattr(self.tensor_db, 'store'):
            self.tensor_db.store = MethodType(_store, self.tensor_db)
        if not hasattr(self.tensor_db, 'retrieve'):
            self.tensor_db.retrieve = MethodType(_retrieve, self.tensor_db)
        if not hasattr(self.tensor_db, 'search'):
            self.tensor_db.search = MethodType(_search, self.tensor_db)

    def __repr__(self) -> str:
        """Representation of the object."""
        with pd.option_context('display.max_rows', None):
            content = self.tensor_db[['tensor_name', 'origin', 'round', 'report', 'tags']]
            return f'TensorDB contents:\n{content}'

    def __str__(self) -> str:
        """Printable string representation."""
        return self.__repr__()

    def clean_up(self, remove_older_than: int = 1) -> None:
        """Remove old entries from database preventing the db from becoming too large and slow."""
        if remove_older_than < 0:
            # Getting a negative argument calls off cleaning
            return
        current_round = self.tensor_db['round'].astype(int).max()
        if current_round == ROUND_PLACEHOLDER:
            current_round = np.sort(self.tensor_db['round'].astype(int).unique())[-2]
        if self.nn:
            self.tensor_db = self.tensor_db[
                (self.tensor_db['round'].astype(int) > current_round - remove_older_than)
            ].reset_index(drop=True)
        else:
            if self.agg:
                self.tensor_db = self.tensor_db[
                    (self.tensor_db['round'].astype(int) > current_round - remove_older_than) |
                    (self.tensor_db['tags'] == ('weak_learner',))
                    ].reset_index(drop=True)
            else:
                self.tensor_db = self.tensor_db[
                    (self.tensor_db['round'].astype(int) > current_round - remove_older_than)
                ].reset_index(drop=True)

    def cache_tensor(self, tensor_key_dict: Dict[TensorKey, np.ndarray]) -> None:
        """Insert tensor into TensorDB (dataframe).

        Args:
            tensor_key_dict: The Tensor Key

        Returns:
            None
        """
        entries_to_add = []
        with self.mutex:
            for tensor_key, nparray in tensor_key_dict.items():
                tensor_name, origin, fl_round, report, tags = tensor_key
                entries_to_add.append(
                    pd.DataFrame([
                        [tensor_name, origin, fl_round, report, tags, nparray]
                    ],
                        columns=[
                            'tensor_name',
                            'origin',
                            'round',
                            'report',
                            'tags',
                            'nparray']
                    )
                )

            self.tensor_db = pd.concat(
                [self.tensor_db, *entries_to_add], ignore_index=True
            )

    def get_tensor_from_cache(self, tensor_key: TensorKey) -> Optional[np.ndarray]:
        """
        Perform a lookup of the tensor_key in the TensorDB.

        Returns the nparray if it is available
        Otherwise, it returns 'None'
        """
        tensor_name, origin, fl_round, report, tags = tensor_key

        # TODO come up with easy way to ignore compression
        df = self.tensor_db[(self.tensor_db['tensor_name'] == tensor_name)
                            & (self.tensor_db['origin'] == origin)
                            & (self.tensor_db['round'] == fl_round)
                            & (self.tensor_db['report'] == report)
                            & (self.tensor_db['tags'] == tags)]

        if len(df) == 0:
            return None

        if self.nn:
            result = np.array(df['nparray'].iloc[0])
        else:
            result = df['nparray'].iloc[0]

        return result

    def get_aggregated_tensor(self, tensor_key: TensorKey, collaborator_weight_dict: dict,
                              aggregation_function: AggregationFunction
                              ) -> Optional[np.ndarray]:
        """
        Determine whether all of the collaborator tensors are present for a given tensor key.

        Returns their weighted average.

        Args:
            tensor_key: The tensor key to be resolved. If origin 'agg_uuid' is
                        present, can be returned directly. Otherwise must
                        compute weighted average of all collaborators
            collaborator_weight_dict: List of collaborator names in federation
                                      and their respective weights
            aggregation_function: Call the underlying numpy aggregation
                                   function. Default is just the weighted
                                   average.
        Returns:
            weighted_nparray if all collaborator values are present
            None if not all values are present

        """
        if len(collaborator_weight_dict) != 0:
            assert np.abs(1.0 - sum(collaborator_weight_dict.values())) < 0.01, (
                f'Collaborator weights do not sum to 1.0: {collaborator_weight_dict}'
            )

        collaborator_names = collaborator_weight_dict.keys()
        agg_tensor_dict = {}

        # Check if the aggregated tensor is already present in TensorDB
        tensor_name, origin, fl_round, report, tags = tensor_key

        raw_df = self.tensor_db[(self.tensor_db['tensor_name'] == tensor_name)
                                & (self.tensor_db['origin'] == origin)
                                & (self.tensor_db['round'] == fl_round)
                                & (self.tensor_db['report'] == report)
                                & (self.tensor_db['tags'] == tags)]['nparray']
        if len(raw_df) > 0:
            return np.array(raw_df.iloc[0]), {}

        for col in collaborator_names:
            new_tags = change_tags(tags, add_field=col)
            raw_df = self.tensor_db[
                (self.tensor_db['tensor_name'] == tensor_name)
                & (self.tensor_db['origin'] == origin)
                & (self.tensor_db['round'] == fl_round)
                & (self.tensor_db['report'] == report)
                & (self.tensor_db['tags'] == new_tags)]['nparray']
            if len(raw_df) == 0:
                tk = TensorKey(tensor_name, origin, report, fl_round, new_tags)
                print(f'No results for collaborator {col}, TensorKey={tk}')
                return None
            else:
                agg_tensor_dict[col] = raw_df.iloc[0]

        local_tensors = [LocalTensor(col_name=col_name,
                                     tensor=agg_tensor_dict[col_name],
                                     weight=collaborator_weight_dict[col_name])
                         for col_name in collaborator_names]

        if hasattr(aggregation_function, '_privileged'):
            if (aggregation_function._privileged):
                with self.mutex:
                    # self.tensor_db.store = MethodType(_store, self.tensor_db)
                    # self.tensor_db.retrieve = MethodType(_retrieve, self.tensor_db)
                    # self.tensor_db.search = MethodType(_search, self.tensor_db)
                    self._bind_convenience_methods()
                    agg_nparray = aggregation_function(local_tensors,
                                                       self.tensor_db,
                                                       tensor_name,
                                                       fl_round,
                                                       tags)
                self.cache_tensor({tensor_key: agg_nparray})

                return np.array(agg_nparray)

        db_iterator = self._iterate()
        agg_nparray = aggregation_function(local_tensors,
                                           db_iterator,
                                           tensor_name,
                                           fl_round,
                                           tags)
        self.cache_tensor({tensor_key: agg_nparray})

        if self.nn or 'metric' in tags:
            result = np.array(agg_nparray)
        else:
            result = agg_nparray

        return result

    # @TODO: this is also to be generalised
    def get_errors(self, round_number):
        df = self.tensor_db[(self.tensor_db['tensor_name'] == "errors")
                            & (self.tensor_db['round'] == round_number)]

        if len(df) == 0:
            return None
        return df["nparray"].to_numpy()

    def _iterate(self, order_by: str = 'round', ascending: bool = False) -> Iterator[pd.Series]:
        columns = ['round', 'nparray', 'tensor_name', 'tags']
        rows = self.tensor_db[columns].sort_values(by=order_by, ascending=ascending).iterrows()
        for _, row in rows:
            yield row



import numpy as np

from openfl.interface.aggregation_functions import AggregationFunction


class AggregateAdaboost(AggregationFunction):
    """Function for Adaboost.F aggregation"""

    def __init__(self, n_classes):
        super(AggregateAdaboost, self).__init__()
        self.n_classes = n_classes

    def call(self, local_tensors, *_):
        tensors = [x.tensor for x in local_tensors]

        errors = np.array([tensor[:-1] for tensor in tensors])
        norm = sum([tensor[-1] for tensor in tensors])
        wl_errs = errors.sum(axis=0) / norm

        best_model = wl_errs.argmin()
        epsilon = wl_errs.min()
        alpha = np.log((1 - epsilon) / (epsilon + 1e-10)) + np.log(self.n_classes - 1)

        return np.array([alpha, best_model])



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import tqdm

from openfl.federated import PyTorchTaskRunner
from openfl.utilities import TensorKey
from .pt_unet_parts import DoubleConv
from .pt_unet_parts import Down
from .pt_unet_parts import soft_dice_coef
from .pt_unet_parts import soft_dice_loss
from .pt_unet_parts import Up


class PyTorchFederatedUnet(PyTorchTaskRunner):
    """Simple Unet for segmentation."""

    def __init__(self, device='cpu', **kwargs):
        """Initialize.

        Args:
            device: The hardware device to use for training (Default = "cpu")
            **kwargs: Additional arguments to pass to the function

        """
        super().__init__(device=device, **kwargs)
        self.init_network(device=self.device, **kwargs)
        self._init_optimizer()
        self.loss_fn = soft_dice_loss
        self.initialize_tensorkeys_for_functions()

    def _init_optimizer(self):
        """Initialize the optimizer."""
        self.optimizer = optim.Adam(self.parameters(), lr=1e-3)

    def init_network(self,
                     device,
                     n_channels,
                     n_classes,
                     print_model=True,
                     **kwargs):
        """Create the network (model).

        Args:
            device: The hardware device to use for training
            n_channels: Number of input image channels
            n_classes: Number of output classes (1 for segmentation)
            print_model: Print the model topology (Default=True)
            **kwargs: Additional arguments to pass to the function

        """
        self.n_channels = n_channels
        self.n_classes = n_classes
        self.inc = DoubleConv(self.n_channels, 64)
        self.down1 = Down(64, 128)
        self.down2 = Down(128, 256)
        self.down3 = Down(256, 512)
        self.down4 = Down(512, 1024)
        self.up1 = Up(1024, 512)
        self.up2 = Up(512, 256)
        self.up3 = Up(256, 128)
        self.up4 = Up(128, 64)
        self.outc = nn.Conv2d(64, self.n_classes, 1)
        if print_model:
            print(self)
        self.to(device)

    def forward(self, x):
        """Forward pass of the model.

        Args:
            x: Data input to the model for the forward pass
        """
        x1 = self.inc(x)
        x2 = self.down1(x1)
        x3 = self.down2(x2)
        x4 = self.down3(x3)
        x5 = self.down4(x4)
        x = self.up1(x5, x4)
        x = self.up2(x, x3)
        x = self.up3(x, x2)
        x = self.up4(x, x1)
        x = self.outc(x)
        x = torch.sigmoid(x)
        return x

    def validate(self, col_name, round_num, input_tensor_dict, use_tqdm=True, **kwargs):
        """Run validation of the model on the local data.

        Args:
            col_name:            Name of the collaborator
            round_num:           What round is it
            input_tensor_dict:   Required input tensors (for model)
            use_tqdm:     Use tqdm to print a progress bar (Default=True)

        Returns:
            global_output_dict:  Tensors to send back to the aggregator
            local_output_dict:   Tensors to maintain in the local TensorDB
        """
        self.rebuild_model(round_num, input_tensor_dict, validation=True)
        self.eval()
        self.to(self.device)
        val_score = 0
        total_samples = 0

        loader = self.data_loader.get_valid_loader()
        if use_tqdm:
            loader = tqdm.tqdm(loader, desc='validate')

        with torch.no_grad():
            for data, target in loader:
                samples = target.shape[0]
                total_samples += samples
                data, target = torch.tensor(data).to(self.device), torch.tensor(
                    target).to(self.device)
                output = self(data)
                # get the index of the max log-probability
                val = soft_dice_coef(output, target)
                val_score += val.sum().cpu().numpy()

        origin = col_name
        suffix = 'validate'
        if kwargs['apply'] == 'local':
            suffix += '_local'
        else:
            suffix += '_agg'
        tags = ('metric', suffix)
        # TODO figure out a better way to pass in metric for this pytorch
        #  validate function
        output_tensor_dict = {
            TensorKey('dice_coef', origin, round_num, True, tags):
                np.array(val_score / total_samples)
        }

        return output_tensor_dict, {}

    def reset_opt_vars(self):
        """Reset optimizer variables.

        Resets the optimizer state variables.

        """
        self._init_optimizer()



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

from openfl.federated import TensorFlowDataLoader
from .tfds_utils import load_histology_shard


class TensorFlowHistologyInMemory(TensorFlowDataLoader):
    """TensorFlow Data Loader for Colorectal Histology Dataset."""

    def __init__(self, data_path, batch_size, **kwargs):
        """
        Initialize.

        Args:
            data_path: File path for the dataset
            batch_size (int): The batch size for the data loader
            **kwargs: Additional arguments, passed to super init and load_mnist_shard
        """
        super().__init__(batch_size, **kwargs)

        _, num_classes, X_train, y_train, X_valid, y_valid = load_histology_shard(
            shard_num=data_path,
            categorical=False, **kwargs
        )

        self.X_train = X_train
        self.y_train = y_train
        self.X_valid = X_valid
        self.y_valid = y_valid

        self.num_classes = num_classes



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""MVTec shard descriptor."""

import os
from glob import glob
from pathlib import Path

import numpy as np
from imageio import imread
from PIL import Image

from openfl.interface.interactive_api.shard_descriptor import ShardDataset
from openfl.interface.interactive_api.shard_descriptor import ShardDescriptor


class MVTecShardDataset(ShardDataset):
    """MVTec Shard dataset class."""

    def __init__(self, images_path,
                 mask_path, labels,
                 rank=1,
                 worldsize=1):
        """Initialize MVTecShardDataset."""
        self.rank = rank
        self.worldsize = worldsize
        self.images_path = images_path[self.rank - 1::self.worldsize]
        self.mask_path = mask_path[self.rank - 1::self.worldsize]
        self.labels = labels[self.rank - 1::self.worldsize]

    def __getitem__(self, index):
        """Return a item by the index."""
        img = np.asarray(imread(self.images_path[index]))
        if img.shape[-1] != 3:
            img = self.gray2rgb(img)

        img = self.resize(img)
        img = np.asarray(img)
        label = self.labels[index]
        if self.mask_path[index]:
            mask = np.asarray(imread(self.mask_path[index]))
            mask = self.resize(mask)
            mask = np.asarray(mask)
        else:
            mask = np.zeros(img.shape)[:, :, 0]
        return img, mask, label

    def __len__(self):
        """Return the len of the dataset."""
        return len(self.images_path)

    def resize(self, image, shape=(256, 256)):
        """Resize image."""
        return np.array(Image.fromarray(image).resize(shape))

    def gray2rgb(self, images):
        """Change image from gray to rgb."""
        tile_shape = tuple(np.ones(len(images.shape), dtype=int))
        tile_shape += (3,)

        images = np.tile(np.expand_dims(images, axis=-1), tile_shape)
        return images


class MVTecShardDescriptor(ShardDescriptor):
    """MVTec Shard descriptor class."""

    def __init__(self, data_folder: str = 'MVTec_data',
                 rank_worldsize: str = '1,1',
                 obj: str = 'bottle'):
        """Initialize MVTecShardDescriptor."""
        super().__init__()

        self.dataset_path = Path.cwd() / data_folder
        self.download_data()
        self.rank, self.worldsize = tuple(int(num) for num in rank_worldsize.split(','))
        self.obj = obj

        # Calculating data and target shapes
        ds = self.get_dataset()
        sample, masks, target = ds[0]
        self._sample_shape = [str(dim) for dim in sample.shape]
        self._target_shape = [str(dim) for dim in target.shape]

    def download_data(self):
        """Download data."""
        zip_file_path = self.dataset_path / 'mvtec_anomaly_detection.tar.xz'
        if not Path(zip_file_path).exists():
            os.makedirs(self.dataset_path, exist_ok=True)
            print('Downloading MVTec Dataset...this might take a while')
            os.system('wget -nc'
                      " 'https://www.mydrive.ch/shares/38536/3830184030e49fe74747669442f0f282/download/420938113-1629952094/mvtec_anomaly_detection.tar.xz'"  # noqa
                      f' -O {zip_file_path.relative_to(Path.cwd())}')
            print('Downloaded MVTec dataset, untar-ring now')
            os.system(f'tar -xvf {zip_file_path.relative_to(Path.cwd())}'
                      f' -C {self.dataset_path.relative_to(Path.cwd())}')
            # change to write permissions
            self.change_permissions(self.dataset_path, 0o764)

    def change_permissions(self, folder, code):
        """Change permissions after data is downloaded."""
        for root, dirs, files in os.walk(folder):
            for d in dirs:
                os.chmod(os.path.join(root, d), code)
            for f in files:
                os.chmod(os.path.join(root, f), code)

    def get_dataset(self, dataset_type='train'):
        """Return a shard dataset by type."""
        # Train dataset
        if dataset_type == 'train':
            fpattern = os.path.join(self.dataset_path, f'{self.obj}/train/*/*.png')
            fpaths = sorted(glob(fpattern))
            self.images_path = list(fpaths)
            self.labels = np.zeros(len(fpaths), dtype=np.int32)
            # Masks
            self.mask_path = np.full(self.labels.shape, None)
        # Test dataset
        elif dataset_type == 'test':
            fpattern = os.path.join(self.dataset_path, f'{self.obj}/test/*/*.png')
            fpaths = sorted(glob(fpattern))
            fpaths_anom = list(
                filter(lambda fpath: os.path.basename(os.path.dirname(fpath)) != 'good', fpaths))
            fpaths_good = list(
                filter(lambda fpath: os.path.basename(os.path.dirname(fpath)) == 'good', fpaths))
            fpaths = fpaths_anom + fpaths_good
            self.images_path = fpaths
            self.labels = np.zeros(len(fpaths_anom) + len(fpaths_good), dtype=np.int32)
            self.labels[:len(fpaths_anom)] = 1   # anomalies
            # Masks
            fpattern_mask = os.path.join(self.dataset_path, f'{self.obj}/ground_truth/*/*.png')
            self.mask_path = sorted(glob(fpattern_mask)) + [None] * len(fpaths_good)
        else:
            raise Exception(f'Wrong dataset type: {dataset_type}.'
                            f'Choose from the list: [train, test]')

        return MVTecShardDataset(
            images_path=self.images_path,
            mask_path=self.mask_path,
            labels=self.labels,
            rank=self.rank,
            worldsize=self.worldsize,
        )

    @property
    def sample_shape(self):
        """Return the sample shape info."""
        return ['256', '256', '3']

    @property
    def target_shape(self):
        """Return the target shape info."""
        return ['256', '256']

    @property
    def dataset_description(self) -> str:
        """Return the shard dataset description."""
        return (f'MVTec dataset, shard number {self.rank}'
                f' out of {self.worldsize}')



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Data transform functions."""

import numpy as np
from sklearn.metrics import balanced_accuracy_score
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import roc_auc_score


def bilinears(images, shape) -> np.ndarray:
    """Generate binlinears."""
    import cv2
    n = images.shape[0]
    new_shape = (n,) + shape
    ret = np.zeros(new_shape, dtype=images.dtype)
    for i in range(n):
        ret[i] = cv2.resize(images[i], dsize=shape[::-1], interpolation=cv2.INTER_LINEAR)
    return ret


def bal_acc_score(obj, predictions, labels):
    """Calculate balanced accuracy score."""
    precision, recall, thresholds = precision_recall_curve(labels.flatten(), predictions.flatten())
    f1_score = (2 * precision * recall) / (precision + recall)
    threshold = thresholds[np.argmax(f1_score)]
    prediction_result = predictions > threshold
    ba_score = balanced_accuracy_score(labels, prediction_result)
    return ba_score


def detection_auroc(obj, anomaly_scores, labels):
    """Calculate detection auroc."""
    # 1: anomaly 0: normal
    auroc = roc_auc_score(labels, anomaly_scores)
    return auroc


def segmentation_auroc(obj, anomaly_maps, masks):
    """Calculate segmentation auroc."""
    gt = masks
    gt = gt.astype(np.int32)
    gt[gt == 255] = 1  # 1: anomaly

    anomaly_maps = bilinears(anomaly_maps, (256, 256))
    auroc = roc_auc_score(gt.flatten(), anomaly_maps.flatten())
    return auroc



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""openfl.transport package."""

from .grpc import AggregatorGRPCClient
from .grpc import AggregatorGRPCServer
from .grpc import DirectorGRPCServer

__all__ = [
    'AggregatorGRPCServer',
    'AggregatorGRPCClient',
    'DirectorGRPCServer',
]



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Tools for metric computation and Dataloader."""

import copy
import random
from collections import defaultdict
from logging import getLogger

import numpy as np
import torch
from torch.utils.data.sampler import Sampler

logger = getLogger(__name__)


class AverageMeter:
    """
    Computes and stores the average and current value.

    Code imported from https://github.com/pytorch/examples/blob/master/imagenet/main.py#L247-L262
    """

    def __init__(self):
        """Initialize Average Meter."""
        self.reset()

    def reset(self):
        """Reset values."""
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        """Update values."""
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count


def compute_ap_cmc(index, good_index, junk_index):
    """Compute validation metrics."""
    ap = 0
    cmc = np.zeros(len(index))

    # remove junk_index
    mask = np.in1d(index, junk_index, invert=True)
    index = index[mask]

    # find good_index index
    ngood = len(good_index)
    mask = np.in1d(index, good_index)
    rows_good = np.argwhere(mask)
    rows_good = rows_good.flatten()

    cmc[rows_good[0]:] = 1.0
    for i in range(ngood):
        d_recall = 1.0 / ngood
        precision = (i + 1) * 1.0 / (rows_good[i] + 1)
        ap = ap + d_recall * precision

    return ap, cmc


def evaluate(distmat, q_pids, g_pids, q_camids, g_camids):
    """Evaluate model."""
    num_q, num_g = distmat.shape
    index = np.argsort(distmat, axis=1)  # from small to large

    num_no_gt = 0  # num of query imgs without groundtruth
    num_r1 = 0
    cmc = np.zeros(len(g_pids))
    ap = 0

    for i in range(num_q):
        # groundtruth index
        query_index = np.argwhere(g_pids == q_pids[i])
        camera_index = np.argwhere(g_camids == q_camids[i])
        good_index = np.setdiff1d(query_index, camera_index, assume_unique=True)
        if good_index.size == 0:
            num_no_gt += 1
            continue
        # remove gallery samples that have the same pid and camid with query
        junk_index = np.intersect1d(query_index, camera_index)

        ap_tmp, cmc_tmp = compute_ap_cmc(index[i], good_index, junk_index)
        if cmc_tmp[0] == 1:
            num_r1 += 1
        cmc = cmc + cmc_tmp
        ap += ap_tmp

    if num_no_gt > 0:
        logger.error(f'{num_no_gt} query imgs do not have groundtruth.')

    cmc = cmc / (num_q - num_no_gt)
    mean_ap = ap / (num_q - num_no_gt)

    return cmc, mean_ap


@torch.no_grad()
def extract_feature(model, dataloader):
    """Extract features for validation."""
    features, pids, camids = [], [], []
    for imgs, (batch_pids, batch_camids) in dataloader:
        flip_imgs = fliplr(imgs)
        imgs, flip_imgs = imgs.cuda(), flip_imgs.cuda()
        batch_features = model(imgs).data
        batch_features_flip = model(flip_imgs).data
        batch_features += batch_features_flip

        features.append(batch_features)
        pids.append(batch_pids)
        camids.append(batch_camids)
    features = torch.cat(features, 0)
    pids = torch.cat(pids, 0).numpy()
    camids = torch.cat(camids, 0).numpy()

    return features, pids, camids


def fliplr(img):
    """Flip horizontal."""
    inv_idx = torch.arange(img.size(3) - 1, -1, -1).long()  # N x C x H x W
    img_flip = img.index_select(3, inv_idx)

    return img_flip


class RandomIdentitySampler(Sampler):
    """
    Random Sampler.

    Randomly sample N identities, then for each identity,
    randomly sample K instances, therefore batch size is N*K.

    Args:
    - data_source (Dataset): dataset to sample from.
    - num_instances (int): number of instances per identity.
    """

    def __init__(self, data_source, num_instances=4):
        """Initialize Sampler."""
        self.data_source = data_source
        self.num_instances = num_instances
        self.index_dic = defaultdict(list)
        for index, (_, (pid, _)) in enumerate(data_source):
            self.index_dic[pid].append(index)
        self.pids = list(self.index_dic.keys())
        self.num_identities = len(self.pids)

        # compute number of examples in an epoch
        self.length = 0
        for pid in self.pids:
            idxs = self.index_dic[pid]
            num = len(idxs)
            if num < self.num_instances:
                num = self.num_instances
            self.length += num - num % self.num_instances

    def __iter__(self):
        """Iterate over Sampler."""
        list_container = []

        for pid in self.pids:
            idxs = copy.deepcopy(self.index_dic[pid])
            if len(idxs) < self.num_instances:
                idxs = np.random.choice(idxs, size=self.num_instances, replace=True)
            random.shuffle(idxs)
            batch_idxs = []
            for idx in idxs:
                batch_idxs.append(idx)
                if len(batch_idxs) == self.num_instances:
                    list_container.append(batch_idxs)
                    batch_idxs = []

        random.shuffle(list_container)

        ret = []
        for batch_idxs in list_container:
            ret.extend(batch_idxs)

        return iter(ret)

    def __len__(self):
        """Return number of examples in an epoch."""
        return self.length



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Federation API module."""

from openfl.transport.grpc.director_client import DirectorClient
from openfl.utilities.utils import getfqdn_env
from .shard_descriptor import DummyShardDescriptor


class Federation:
    """
    Federation class.

    Federation entity exists to keep information about collaborator related settings,
    their local data and network setting to enable communication in federation.
    """

    def __init__(self, client_id=None, director_node_fqdn=None, director_port=None, tls=True,
                 cert_chain=None, api_cert=None, api_private_key=None) -> None:
        """
        Initialize federation.

        Federation API class should be initialized with the Director node FQDN
        and encryption settings. One may disable mTLS in trusted environments or
        provide paths to a certificate chain to CA, API certificate and
        pricate key to enable mTLS.

        Args:
        - client_id: name of created Frontend API instance.
            The same name user certify.
        - director_node_fqdn: Address and port a director's service is running on.
            User passes here an address with a port.
        """
        if director_node_fqdn is None:
            self.director_node_fqdn = getfqdn_env()
        else:
            self.director_node_fqdn = director_node_fqdn

        self.tls = tls

        self.cert_chain = cert_chain
        self.api_cert = api_cert
        self.api_private_key = api_private_key

        # Create Director client
        self.dir_client = DirectorClient(
            client_id=client_id,
            director_host=director_node_fqdn,
            director_port=director_port,
            tls=tls,
            root_certificate=cert_chain,
            private_key=api_private_key,
            certificate=api_cert
        )

        # Request sample and target shapes from Director.
        # This is an internal method for finding out dataset properties in a Federation.
        self.sample_shape, self.target_shape = self.dir_client.get_dataset_info()

    def get_dummy_shard_descriptor(self, size):
        """Return a dummy shard descriptor."""
        return DummyShardDescriptor(self.sample_shape, self.target_shape, size)

    def get_shard_registry(self):
        """Return a shard registry."""
        return self.dir_client.get_envoys()



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""gRPC interceptors module."""
import collections

import grpc


class _GenericClientInterceptor(grpc.UnaryUnaryClientInterceptor,
                                grpc.UnaryStreamClientInterceptor,
                                grpc.StreamUnaryClientInterceptor,
                                grpc.StreamStreamClientInterceptor):

    def __init__(self, interceptor_function):
        self._fn = interceptor_function

    def intercept_unary_unary(self, continuation, client_call_details, request):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, False)
        response = continuation(new_details, next(new_request_iterator))
        return postprocess(response) if postprocess else response

    def intercept_unary_stream(self, continuation, client_call_details,
                               request):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, iter((request,)), False, True)
        response_it = continuation(new_details, next(new_request_iterator))
        return postprocess(response_it) if postprocess else response_it

    def intercept_stream_unary(self, continuation, client_call_details,
                               request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator, True, False)
        response = continuation(new_details, new_request_iterator)
        return postprocess(response) if postprocess else response

    def intercept_stream_stream(self, continuation, client_call_details,
                                request_iterator):
        new_details, new_request_iterator, postprocess = self._fn(
            client_call_details, request_iterator, True, True)
        response_it = continuation(new_details, new_request_iterator)
        return postprocess(response_it) if postprocess else response_it


def _create_generic_interceptor(intercept_call):
    return _GenericClientInterceptor(intercept_call)


class _ClientCallDetails(
    collections.namedtuple(
        '_ClientCallDetails',
        ('method', 'timeout', 'metadata', 'credentials')
    ),
    grpc.ClientCallDetails
):
    pass


def headers_adder(headers):
    """Create interceptor with added headers."""

    def intercept_call(client_call_details, request_iterator, request_streaming,
                       response_streaming):
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        for header, value in headers.items():
            metadata.append((
                header,
                value,
            ))
        client_call_details = _ClientCallDetails(
            client_call_details.method, client_call_details.timeout, metadata,
            client_call_details.credentials)
        return client_call_details, request_iterator, None

    return _create_generic_interceptor(intercept_call)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Device monitor plugin module."""


class DeviceMonitor:
    """Device monitor plugin interface."""

    def get_driver_version(self) -> str:
        """Get device's driver version."""
        raise NotImplementedError

    def get_device_utilization(self, index: int) -> str:
        """
        Get device utilization method.

        It is just a general method that returns a string that may be shown to the frontend user.
        """
        raise NotImplementedError



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0-

"""openfl.utilities.data package."""
from openfl.utilities.data_splitters.data_splitter import DataSplitter
from openfl.utilities.data_splitters.numpy import DirichletNumPyDataSplitter
from openfl.utilities.data_splitters.numpy import EqualNumPyDataSplitter
from openfl.utilities.data_splitters.numpy import LogNormalNumPyDataSplitter
from openfl.utilities.data_splitters.numpy import NumPyDataSplitter
from openfl.utilities.data_splitters.numpy import RandomNumPyDataSplitter

__all__ = [
    'DataSplitter',
    'DirichletNumPyDataSplitter',
    'EqualNumPyDataSplitter',
    'LogNormalNumPyDataSplitter',
    'NumPyDataSplitter',
    'RandomNumPyDataSplitter',
]



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""AggregatorGRPCClient module."""

import time
from logging import getLogger
from typing import Optional
from typing import Tuple

import grpc

from openfl.pipelines import NoCompressionPipeline
from openfl.protocols import aggregator_pb2
from openfl.protocols import aggregator_pb2_grpc
from openfl.protocols import utils
from openfl.utilities import check_equal
from .grpc_channel_options import channel_options


class ConstantBackoff:
    """Constant Backoff policy."""

    def __init__(self, reconnect_interval, logger, uri):
        """Initialize Constant Backoff."""
        self.reconnect_interval = reconnect_interval
        self.logger = logger
        self.uri = uri

    def sleep(self):
        """Sleep for specified interval."""
        self.logger.info(f'Attempting to connect to aggregator at {self.uri}')
        time.sleep(self.reconnect_interval)


class RetryOnRpcErrorClientInterceptor(
    grpc.UnaryUnaryClientInterceptor, grpc.StreamUnaryClientInterceptor
):
    """Retry gRPC connection on failure."""

    def __init__(
            self,
            sleeping_policy,
            status_for_retry: Optional[Tuple[grpc.StatusCode]] = None,
    ):
        """Initialize function for gRPC retry."""
        self.sleeping_policy = sleeping_policy
        self.status_for_retry = status_for_retry

    def _intercept_call(self, continuation, client_call_details, request_or_iterator):
        """Intercept the call to the gRPC server."""
        while True:
            response = continuation(client_call_details, request_or_iterator)

            if isinstance(response, grpc.RpcError):

                # If status code is not in retryable status codes
                self.sleeping_policy.logger.info(f'Response code: {response.code()}')
                if (
                        self.status_for_retry
                        and response.code() not in self.status_for_retry
                ):
                    return response

                self.sleeping_policy.sleep()
            else:
                return response

    def intercept_unary_unary(self, continuation, client_call_details, request):
        """Wrap intercept call for unary->unary RPC."""
        return self._intercept_call(continuation, client_call_details, request)

    def intercept_stream_unary(
            self, continuation, client_call_details, request_iterator
    ):
        """Wrap intercept call for stream->unary RPC."""
        return self._intercept_call(continuation, client_call_details, request_iterator)


def _atomic_connection(func):
    def wrapper(self, *args, **kwargs):
        self.reconnect()
        response = func(self, *args, **kwargs)
        self.disconnect()
        return response

    return wrapper


def _resend_data_on_reconnection(func):
    def wrapper(self, *args, **kwargs):
        while True:
            try:
                response = func(self, *args, **kwargs)
            except grpc.RpcError as e:
                if e.code() == grpc.StatusCode.UNKNOWN:
                    self.logger.info(
                        f'Attempting to resend data request to aggregator at {self.uri}'
                    )
                continue
            break
        return response

    return wrapper


class AggregatorGRPCClient:
    """Client to the aggregator over gRPC-TLS."""

    def __init__(self,
                 agg_addr,
                 agg_port,
                 tls,
                 disable_client_auth,
                 root_certificate,
                 certificate,
                 private_key,
                 aggregator_uuid=None,
                 federation_uuid=None,
                 single_col_cert_common_name=None,
                 **kwargs):
        """Initialize."""
        self.uri = f'{agg_addr}:{agg_port}'
        self.tls = tls
        self.disable_client_auth = disable_client_auth
        self.root_certificate = root_certificate
        self.certificate = certificate
        self.private_key = private_key

        self.logger = getLogger(__name__)

        if not self.tls:
            self.logger.warn(
                'gRPC is running on insecure channel with TLS disabled.')
            self.channel = self.create_insecure_channel(self.uri)
        else:
            self.channel = self.create_tls_channel(
                self.uri,
                self.root_certificate,
                self.disable_client_auth,
                self.certificate,
                self.private_key
            )

        self.header = None
        self.aggregator_uuid = aggregator_uuid
        self.federation_uuid = federation_uuid
        self.single_col_cert_common_name = single_col_cert_common_name

        # Adding an interceptor for RPC Errors
        self.interceptors = (
            RetryOnRpcErrorClientInterceptor(
                sleeping_policy=ConstantBackoff(
                    logger=self.logger,
                    reconnect_interval=int(kwargs.get('client_reconnect_interval', 1)),
                    uri=self.uri),
                status_for_retry=(grpc.StatusCode.UNAVAILABLE,),
            ),
        )
        self.stub = aggregator_pb2_grpc.AggregatorStub(
            grpc.intercept_channel(self.channel, *self.interceptors)
        )

    def create_insecure_channel(self, uri):
        """
        Set an insecure gRPC channel (i.e. no TLS) if desired.

        Warns user that this is not recommended.

        Args:
            uri: The uniform resource identifier fo the insecure channel

        Returns:
            An insecure gRPC channel object

        """
        return grpc.insecure_channel(uri, options=channel_options)

    def create_tls_channel(self, uri, root_certificate, disable_client_auth,
                           certificate, private_key):
        """
        Set an secure gRPC channel (i.e. TLS).

        Args:
            uri: The uniform resource identifier fo the insecure channel
            root_certificate: The Certificate Authority filename
            disable_client_auth (boolean): True disabled client-side
             authentication (not recommended, throws warning to user)
            certificate: The client certficate filename from the collaborator
             (signed by the certificate authority)

        Returns:
            An insecure gRPC channel object
        """
        with open(root_certificate, 'rb') as f:
            root_certificate_b = f.read()

        if disable_client_auth:
            self.logger.warn('Client-side authentication is disabled.')
            private_key_b = None
            certificate_b = None
        else:
            with open(private_key, 'rb') as f:
                private_key_b = f.read()
            with open(certificate, 'rb') as f:
                certificate_b = f.read()

        credentials = grpc.ssl_channel_credentials(
            root_certificates=root_certificate_b,
            private_key=private_key_b,
            certificate_chain=certificate_b,
        )

        return grpc.secure_channel(
            uri, credentials, options=channel_options)

    def _set_header(self, collaborator_name):
        self.header = aggregator_pb2.MessageHeader(
            sender=collaborator_name,
            receiver=self.aggregator_uuid,
            federation_uuid=self.federation_uuid,
            single_col_cert_common_name=self.single_col_cert_common_name or ''
        )

    def validate_response(self, reply, collaborator_name):
        """Validate the aggregator response."""
        # check that the message was intended to go to this collaborator
        check_equal(reply.header.receiver, collaborator_name, self.logger)
        check_equal(reply.header.sender, self.aggregator_uuid, self.logger)

        # check that federation id matches
        check_equal(
            reply.header.federation_uuid,
            self.federation_uuid,
            self.logger
        )

        # check that there is aggrement on the single_col_cert_common_name
        check_equal(
            reply.header.single_col_cert_common_name,
            self.single_col_cert_common_name or '',
            self.logger
        )

    def disconnect(self):
        """Close the gRPC channel."""
        self.logger.debug(f'Disconnecting from gRPC server at {self.uri}')
        self.channel.close()

    def reconnect(self):
        """Create a new channel with the gRPC server."""
        # channel.close() is idempotent. Call again here in case it wasn't issued previously
        self.disconnect()

        if not self.tls:
            self.channel = self.create_insecure_channel(self.uri)
        else:
            self.channel = self.create_tls_channel(
                self.uri,
                self.root_certificate,
                self.disable_client_auth,
                self.certificate,
                self.private_key
            )

        self.logger.debug(f'Connecting to gRPC at {self.uri}')

        self.stub = aggregator_pb2_grpc.AggregatorStub(
            grpc.intercept_channel(self.channel, *self.interceptors)
        )

    @_atomic_connection
    @_resend_data_on_reconnection
    def get_tasks(self, collaborator_name):
        """Get tasks from the aggregator."""
        self._set_header(collaborator_name)
        request = aggregator_pb2.GetTasksRequest(header=self.header)
        response = self.stub.GetTasks(request)
        self.validate_response(response, collaborator_name)

        return response.tasks, response.round_number, response.sleep_time, response.quit

    @_atomic_connection
    @_resend_data_on_reconnection
    def send_local_task_results(self, collaborator_name, round_number,
                                task_name, data_size, named_tensors):
        """Send task results to the aggregator."""
        self._set_header(collaborator_name)
        request = aggregator_pb2.TaskResults(
            header=self.header,
            round_number=round_number,
            task_name=task_name,
            data_size=data_size,
            tensors=named_tensors
        )

        # convert (potentially) long list of tensors into stream
        stream = []
        stream += utils.proto_to_datastream(request, self.logger)
        response = self.stub.SendLocalTaskResults(iter(stream))

        # also do other validation, like on the round_number
        self.validate_response(response, collaborator_name)

    @_atomic_connection
    @_resend_data_on_reconnection
    def get_tensor(self, collaborator_name, tensor_name, round_number,
                   report, tags, require_lossless, aggregated):
        """Get tensor from the aggregator."""
        self._set_header(collaborator_name)
        request = aggregator_pb2.GetTensorRequest(
            header=self.header,
            tensor_name=tensor_name,
            round_number=round_number,
            report=report,
            tags=tags,
            require_lossless=require_lossless,
            aggregated=aggregated
        )
        response = self.stub.GetTensor(request)
        # also do other validation, like on the round_number
        self.validate_response(response, collaborator_name)

        return response.tensor

    @_atomic_connection
    @_resend_data_on_reconnection
    def synch(self, task_name, round_number, collaborator_name):
        """Get tasks from the aggregator."""
        self._set_header(collaborator_name)
        request = aggregator_pb2.SynchRequest(header=self.header, task_name=task_name, round_number=round_number)
        response = self.stub.GetSynch(request)
        self.validate_response(response, collaborator_name)

        return response.is_completed

    def _get_trained_model(self, experiment_name, model_type):
        """Get trained model RPC."""
        get_model_request = self.stub.GetTrainedModelRequest(
            experiment_name=experiment_name,
            model_type=model_type,
        )
        model_proto_response = self.stub.GetTrainedModel(get_model_request)
        tensor_dict, _ = utils.deconstruct_model_proto(
            model_proto_response.model_proto,
            NoCompressionPipeline(),
        )
        return tensor_dict



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""PyTorchDataLoader module."""

from math import ceil

import numpy as np
from openfl.federated.data.loader import DataLoader


class GenericDataLoader(DataLoader):
    """Federation Data Loader for TensorFlow Models."""

    def __init__(self, batch_size, random_seed=None, **kwargs):
        """
        Instantiate the data object.

        Args:
            batch_size: Size of batches used for all data loaders
            kwargs: consumes all un-used kwargs

        Returns:
            None
        """
        super().__init__(**kwargs)
        self.batch_size = batch_size
        self.X_train = kwargs['X_train']
        self.y_train = kwargs['y_train']
        self.X_valid = kwargs['X_test']
        self.y_valid = kwargs['y_test']
        self.random_seed = random_seed

        # Child classes should have init signature:
        # (self, batch_size, **kwargs), should call this __init__ and then
        # define self.X_train, self.y_train, self.X_valid, and self.y_valid

    def get_feature_shape(self):
        """Get the shape of an example feature array.

        Returns:
            tuple: shape of an example feature array
        """
        return len(self.X_train[0])

    def get_train_loader(self, batch_size=None, num_batches=None):
        """
        Get training data loader.

        Returns
        -------
        loader object
        """
        return self._get_batch_generator(
            X=self.X_train, y=self.y_train, batch_size=batch_size, num_batches=num_batches)

    def get_valid_loader(self, batch_size=None):
        """
        Get validation data loader.

        Returns:
            loader object
        """
        return self._get_batch_generator(X=self.X_valid, y=self.y_valid, batch_size=batch_size)

    def get_train_data_size(self):
        """
        Get total number of training samples.

        Returns:
            int: number of training samples
        """
        return len(self.X_train)

    def get_valid_data_size(self):
        """
        Get total number of validation samples.

        Returns:
            int: number of validation samples
        """
        return len(self.X_valid)

    @staticmethod
    def _batch_generator(X, y, idxs, batch_size, num_batches):
        """
        Generate batch of data.

        Args:
            X: input data
            y: label data
            idxs: The index of the dataset
            batch_size: The batch size for the data loader
            num_batches: The number of batches

        Yields:
            tuple: input data, label data

        """
        for i in range(num_batches):
            a = i * batch_size
            b = a + batch_size
            yield X[idxs[a:b]], y[idxs[a:b]]

    def _get_batch_generator(self, X, y, batch_size, num_batches=None):
        """
        Return the dataset generator.

        Args:
            X: input data
            y: label data
            batch_size: The batch size for the data loader

        """
        if batch_size is None:
            batch_size = self.batch_size

        # shuffle data indices
        if self.random_seed is not None:
            np.random.seed(self.random_seed)

        idxs = np.random.permutation(np.arange(len(X.shape)))

        # compute the number of batches
        if num_batches is None:
            num_batches = ceil(len(X.shape) / batch_size)

        # build the generator and return it
        return self._batch_generator(X, y, idxs, batch_size, num_batches)


# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Aggregation function interface module."""
from abc import abstractmethod
from typing import Iterator
from typing import List
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd

from openfl.utilities import LocalTensor
from openfl.utilities import SingletonABCMeta


class AggregationFunction(metaclass=SingletonABCMeta):
    """Interface for specifying aggregation function."""

    def __init__(self):
        """Initialize common AggregationFunction params
        
           Default: Read only access to TensorDB
        """
        self._privileged = False

    @abstractmethod
    def call(self,
             local_tensors: List[LocalTensor],
             db_iterator: Iterator[pd.Series],
             tensor_name: str,
             fl_round: int,
             tags: Tuple[str]) -> np.ndarray:
        """Aggregate tensors.

        Args:
            local_tensors(list[openfl.utilities.LocalTensor]): List of local tensors to aggregate.
            db_iterator: An iterator over history of all tensors. Columns:
                - 'tensor_name': name of the tensor.
                    Examples for `torch.nn.Module`s: 'conv1.weight', 'fc2.bias'.
                - 'round': 0-based number of round corresponding to this tensor.
                - 'tags': tuple of tensor tags. Tags that can appear:
                    - 'model' indicates that the tensor is a model parameter.
                    - 'trained' indicates that tensor is a part of a training result.
                        These tensors are passed to the aggregator node after local learning.
                    - 'aggregated' indicates that tensor is a result of aggregation.
                        These tensors are sent to collaborators for the next round.
                    - 'delta' indicates that value is a difference between rounds
                        for a specific tensor.
                    also one of the tags is a collaborator name
                    if it corresponds to a result of a local task.

                - 'nparray': value of the tensor.
            tensor_name: name of the tensor
            fl_round: round number
            tags: tuple of tags for this tensor
        Returns:
            np.ndarray: aggregated tensor
        """
        raise NotImplementedError

    def __call__(self, local_tensors,
                 db_iterator,
                 tensor_name,
                 fl_round,
                 tags):
        """Use magic function for ease."""
        return self.call(local_tensors, db_iterator, tensor_name, fl_round, tags)




# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""openfl.utilities package."""

from .types import *  # NOQA
from .checks import *  # NOQA
from .utils import *  # NOQA



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
"""Proto utils."""

from openfl.protocols import base_pb2
from openfl.utilities import TensorKey


def model_proto_to_bytes_and_metadata(model_proto, nn):
    """Convert the model protobuf to bytes and metadata.

    Args:
        model_proto: Protobuf of the model

    Returns:
        bytes_dict: Dictionary of the bytes contained in the model protobuf
        metadata_dict: Dictionary of the meta data in the model protobuf
    """
    bytes_dict = {}
    metadata_dict = {}
    round_number = None
    for tensor_proto in model_proto.tensors:
        bytes_dict[tensor_proto.name] = tensor_proto.data_bytes
        metadata_dict[tensor_proto.name] = [{
            'int_to_float': proto.int_to_float,
            'int_list': proto.int_list,
            'bool_list': proto.bool_list,
            'model': not nn
        }
            for proto in tensor_proto.transformer_metadata
        ]
        if round_number is None:
            round_number = tensor_proto.round_number
        else:
            assert round_number == tensor_proto.round_number, (
                f'Round numbers in model are inconsistent: {round_number} '
                f'and {tensor_proto.round_number}'
            )
    return bytes_dict, metadata_dict, round_number


def bytes_and_metadata_to_model_proto(bytes_dict, model_id, model_version,
                                      is_delta, metadata_dict):
    """Convert bytes and metadata to model protobuf."""
    model_header = ModelHeader(id=model_id, version=model_version, is_delta=is_delta)  # NOQA:F821

    tensor_protos = []
    for key, data_bytes in bytes_dict.items():
        transformer_metadata = metadata_dict[key]
        metadata_protos = []
        for metadata in transformer_metadata:
            if metadata.get('int_to_float') is not None:
                int_to_float = metadata.get('int_to_float')
            else:
                int_to_float = {}

            if metadata.get('int_list') is not None:
                int_list = metadata.get('int_list')
            else:
                int_list = []

            if metadata.get('bool_list') is not None:
                bool_list = metadata.get('bool_list')
            else:
                bool_list = []
            metadata_protos.append(base_pb2.MetadataProto(
                int_to_float=int_to_float,
                int_list=int_list,
                bool_list=bool_list,
            ))
        tensor_protos.append(TensorProto(name=key,  # NOQA:F821
                                         data_bytes=data_bytes,
                                         transformer_metadata=metadata_protos))
    return base_pb2.ModelProto(header=model_header, tensors=tensor_protos)


def construct_named_tensor(tensor_key, nparray, transformer_metadata, lossless):
    """Construct named tensor."""
    metadata_protos = []
    for metadata in transformer_metadata:
        if metadata.get('int_to_float') is not None:
            int_to_float = metadata.get('int_to_float')
        else:
            int_to_float = {}

        if metadata.get('int_list') is not None:
            int_list = metadata.get('int_list')
        else:
            int_list = []

        if metadata.get('bool_list') is not None:
            bool_list = metadata.get('bool_list')
        else:
            bool_list = []

        if metadata.get('model') is not None:
            model = metadata.get('model')
        else:
            model = False
        metadata_protos.append(base_pb2.MetadataProto(
            int_to_float=int_to_float,
            int_list=int_list,
            bool_list=bool_list,
            model=model
        ))

    tensor_name, origin, round_number, report, tags = tensor_key

    return base_pb2.NamedTensor(
        name=tensor_name,
        round_number=round_number,
        lossless=lossless,
        report=report,
        tags=tags,
        transformer_metadata=metadata_protos,
        data_bytes=nparray,
    )


def construct_proto(tensor_dict, model_id, model_version, is_delta, compression_pipeline):
    """Construct proto."""
    # compress the arrays in the tensor_dict, and form the model proto
    # TODO: Hold-out tensors from the compression pipeline.
    bytes_dict = {}
    metadata_dict = {}
    for key, array in tensor_dict.items():
        bytes_dict[key], metadata_dict[key] = compression_pipeline.forward(data=array)

    # convert the compressed_tensor_dict and metadata to protobuf, and make the new model proto
    model_proto = bytes_and_metadata_to_model_proto(bytes_dict=bytes_dict,
                                                    model_id=model_id,
                                                    model_version=model_version,
                                                    is_delta=is_delta,
                                                    metadata_dict=metadata_dict)
    return model_proto


def construct_model_proto(tensor_dict, round_number, tensor_pipe):
    """Construct model proto from tensor dict."""
    # compress the arrays in the tensor_dict, and form the model proto
    # TODO: Hold-out tensors from the tensor compression pipeline.
    named_tensors = []
    for key, nparray in tensor_dict.items():
        bytes_data, transformer_metadata = tensor_pipe.forward(data=nparray, key=key)
        tensor_key = TensorKey(key, 'agg', round_number, False, ('model',))
        named_tensors.append(construct_named_tensor(
            tensor_key,
            bytes_data,
            transformer_metadata,
            lossless=True,
        ))

    return base_pb2.ModelProto(tensors=named_tensors)


def deconstruct_model_proto(model_proto, compression_pipeline):
    """Deconstruct model proto."""
    # extract the tensor_dict and metadata
    bytes_dict, metadata_dict, round_number = model_proto_to_bytes_and_metadata(model_proto, compression_pipeline.is_nn())

    # decompress the tensors
    # TODO: Handle tensors meant to be held-out from the compression pipeline
    #  (currently none are held out).
    tensor_dict = {}
    for key in bytes_dict:
        tensor_dict[key] = compression_pipeline.backward(data=bytes_dict[key],
                                                         transformer_metadata=metadata_dict[key])
    return tensor_dict, round_number


def deconstruct_proto(model_proto, compression_pipeline):
    """Deconstruct the protobuf.

    Args:
        model_proto: The protobuf of the model
        compression_pipeline: The compression pipeline object

    Returns:
        protobuf: A protobuf of the model
    """
    # extract the tensor_dict and metadata
    bytes_dict, metadata_dict = model_proto_to_bytes_and_metadata(model_proto, compression_pipeline.is_nn())

    # decompress the tensors
    # TODO: Handle tensors meant to be held-out from the compression pipeline
    #  (currently none are held out).
    tensor_dict = {}
    for key in bytes_dict:
        tensor_dict[key] = compression_pipeline.backward(data=bytes_dict[key],
                                                         transformer_metadata=metadata_dict[key])
    return tensor_dict


def load_proto(fpath):
    """Load the protobuf.

    Args:
        fpath: The filepath for the protobuf

    Returns:
        protobuf: A protobuf of the model
    """
    with open(fpath, 'rb') as f:
        loaded = f.read()
        model = base_pb2.ModelProto().FromString(loaded)
        return model


def dump_proto(model_proto, fpath):
    """Dump the protobuf to a file.

    Args:
        model_proto: The protobuf of the model
        fpath: The filename to save the model protobuf

    """
    s = model_proto.SerializeToString()
    with open(fpath, 'wb') as f:
        f.write(s)


def datastream_to_proto(proto, stream, logger=None):
    """Convert the datastream to the protobuf.

    Args:
        model_proto: The protobuf of the model
        stream: The data stream from the remote connection
        logger: (Optional) The log object

    Returns:
        protobuf: A protobuf of the model
    """
    npbytes = b''
    for chunk in stream:
        npbytes += chunk.npbytes

    if len(npbytes) > 0:
        proto.ParseFromString(npbytes)
        if logger is not None:
            logger.debug(f'datastream_to_proto parsed a {type(proto)}.')
        return proto
    else:
        raise RuntimeError(f'Received empty stream message of type {type(proto)}')


def proto_to_datastream(proto, logger, max_buffer_size=(2 * 1024 * 1024)):
    """Convert the protobuf to the datastream for the remote connection.

    Args:
        model_proto: The protobuf of the model
        logger: The log object
        max_buffer_size: The buffer size (Default= 2*1024*1024)
    Returns:
        reply: The message for the remote connection.
    """
    npbytes = proto.SerializeToString()
    data_size = len(npbytes)
    buffer_size = data_size if max_buffer_size > data_size else max_buffer_size
    logger.debug(f'Setting stream chunks with size {buffer_size} for proto of type {type(proto)}')

    for i in range(0, data_size, buffer_size):
        chunk = npbytes[i: i + buffer_size]
        reply = base_pb2.DataStream(npbytes=chunk, size=len(chunk))
        yield reply


def get_headers(context) -> dict:
    """Get headers from context."""
    return {header[0]: header[1] for header in context.invocation_metadata()}



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import fastestimator as fe
from fastestimator.dataset.data import cifar10
from fastestimator.op.numpyop.univariate import Normalize

from openfl.federated import FastEstimatorDataLoader


class FastEstimatorCifarInMemory(FastEstimatorDataLoader):
    """TensorFlow Data Loader for MNIST Dataset."""

    def __init__(self, data_path, batch_size, **kwargs):
        """
        Initialize.

        Args:
            data_path: File path for the dataset
            batch_size (int): The batch size for the data loader
            **kwargs: Additional arguments, passed to super init and
             load_mnist_shard
        """
        # TODO: We should be downloading the dataset shard into a directory
        # TODO: There needs to be a method to ask how many collaborators and
        #  what index/rank is this collaborator.
        # Then we have a way to automatically shard based on rank and size
        # of collaborator list.

        train_data, eval_data = cifar10.load_data()
        test_data = eval_data.split(0.5)

        collaborator_count = kwargs['collaborator_count']

        train_data, eval_data, test_data = self.split_data(
            train_data,
            eval_data,
            test_data,
            int(data_path),
            collaborator_count
        )

        print(f'train_data = {train_data}')
        print(f'eval_data = {eval_data}')
        print(f'test_data = {test_data}')

        print(f'batch_size = {batch_size}')

        super().__init__(fe.Pipeline(
            train_data=train_data,
            eval_data=eval_data,
            test_data=test_data,
            batch_size=batch_size,
            ops=[
                Normalize(inputs='x', outputs='x',
                          mean=(0.4914, 0.4822, 0.4465),
                          std=(0.2471, 0.2435, 0.2616))
            ]), **kwargs)

    def split_data(self, train, eva, test, rank, collaborator_count):
        """Split data into N parts, where N is the collaborator count."""
        if collaborator_count == 1:
            return train, eva, test

        fraction = [1.0 / float(collaborator_count)]
        fraction *= (collaborator_count - 1)

        # Expand the split list into individual parameters
        train_split = train.split(*fraction)
        eva_split = eva.split(*fraction)
        test_split = test.split(*fraction)

        train = [train]
        eva = [eva]
        test = [test]

        if type(train_split) is not list:
            train.append(train_split)
            eva.append(eva_split)
            test.append(test_split)
        else:
            # Combine all partitions into a single list
            train = [train] + train_split
            eva = [eva] + eva_split
            test = [test] + test_split

        # Extract the right shard
        train = train[rank - 1]
        eva = eva[rank - 1]
        test = test[rank - 1]

        return train, eva, test



from sklearn.ensemble import RandomForestClassifier


class MyRandomForestClassifier(RandomForestClassifier):
    def __init__(self):
        super(MyRandomForestClassifier, self).__init__(n_estimators=1, max_depth=2)



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Inspection of images and patches."""

import os
import shutil

import ngtpy
import numpy as np
from data_transf import bal_acc_score
from data_transf import detection_auroc
from data_transf import segmentation_auroc
from sklearn.neighbors import KDTree
from utils import distribute_scores


def search_nn(test_emb, train_emb_flat, nn=1, method='kdt'):
    """Seach nearest neighbors."""
    if method == 'ngt':
        return search_nn_ngt(test_emb, train_emb_flat, nn=nn)

    kdt = KDTree(train_emb_flat)

    ntest, i, j, d = test_emb.shape
    closest_inds = np.empty((ntest, i, j, nn), dtype=np.int32)
    l2_maps = np.empty((ntest, i, j, nn), dtype=np.float32)

    for n_ in range(ntest):
        for i_ in range(i):
            dists, inds = kdt.query(test_emb[n_, i_, :, :], return_distance=True, k=nn)
            closest_inds[n_, i_, :, :] = inds[:, :]
            l2_maps[n_, i_, :, :] = dists[:, :]

    return l2_maps, closest_inds


def search_nn_ngt(test_emb, train_emb_flat, nn=1):
    """Search nearest neighbors."""
    ntest, i, j, d = test_emb.shape
    closest_inds = np.empty((ntest, i, j, nn), dtype=np.int32)
    l2_maps = np.empty((ntest, i, j, nn), dtype=np.float32)

    dpath = f'/tmp/{os.getpid()}'
    ngtpy.create(dpath, d)
    index = ngtpy.Index(dpath)
    index.batch_insert(train_emb_flat)

    for n_ in range(ntest):
        for i_ in range(i):
            for j_ in range(j):
                query = test_emb[n_, i_, j_, :]
                results = index.search(query, nn)
                inds = [result[0] for result in results]

                closest_inds[n_, i_, j_, :] = inds
                vecs = np.asarray([index.get_object(inds[nn_]) for nn_ in range(nn)])
                dists = np.linalg.norm(query - vecs, axis=-1)
                l2_maps[n_, i_, j_, :] = dists
    shutil.rmtree(dpath)

    return l2_maps, closest_inds


def assess_anomaly_maps(obj, anomaly_maps, masks, labels):
    """Assess anomaly maps."""
    auroc_seg = segmentation_auroc(obj, anomaly_maps, masks)

    anomaly_scores = anomaly_maps.max(axis=-1).max(axis=-1)
    auroc_det = detection_auroc(obj, anomaly_scores, labels)
    ba_score = bal_acc_score(obj, anomaly_scores, labels)
    return auroc_det, auroc_seg, ba_score


def eval_embeddings_nn_multik(obj, embs64, embs32, masks, labels, nn=1):
    """Evaluate embeddings."""
    emb_tr, emb_te = embs64
    maps_64 = measure_emb_nn(emb_te, emb_tr, method='kdt', nn=nn)
    maps_64 = distribute_scores(maps_64, (256, 256), k=64, s=16)
    det_64, seg_64, ba_64 = assess_anomaly_maps(obj, maps_64, masks, labels)

    emb_tr, emb_te = embs32
    maps_32 = measure_emb_nn(emb_te, emb_tr, method='ngt', nn=nn)
    maps_32 = distribute_scores(maps_32, (256, 256), k=32, s=4)
    det_32, seg_32, ba_32 = assess_anomaly_maps(obj, maps_32, masks, labels)

    maps_sum = maps_64 + maps_32
    det_sum, seg_sum, ba_sum = assess_anomaly_maps(obj, maps_sum, masks, labels)

    maps_mult = maps_64 * maps_32
    det_mult, seg_mult, ba_mult = assess_anomaly_maps(obj, maps_mult, masks, labels)

    return {
        'det_64': det_64,
        'seg_64': seg_64,
        'bal_acc_64': ba_64,

        'det_32': det_32,
        'seg_32': seg_32,
        'bal_acc_32': ba_32,

        'det_sum': det_sum,
        'seg_sum': seg_sum,
        'bal_acc_sum': ba_sum,

        'det_mult': det_mult,
        'seg_mult': seg_mult,
        'bal_acc_mult': ba_mult,

        'maps_64': maps_64,
        'maps_32': maps_32,
        'maps_sum': maps_sum,
        'maps_mult': maps_mult,
    }


def eval_embeddings_nn_maps(obj, embs64, embs32, masks, labels, nn=1):
    """Evaluate embeddings."""
    emb_tr, emb_te = embs64
    maps_64 = measure_emb_nn(emb_te, emb_tr, method='kdt', nn=nn)
    maps_64 = distribute_scores(maps_64, (256, 256), k=64, s=16)

    emb_tr, emb_te = embs32
    maps_32 = measure_emb_nn(emb_te, emb_tr, method='ngt', nn=nn)
    maps_32 = distribute_scores(maps_32, (256, 256), k=32, s=4)

    maps_mult = maps_64 * maps_32
    return maps_mult


def measure_emb_nn(emb_te, emb_tr, method='kdt', nn=1):
    """Measure embeddings."""
    d = emb_tr.shape[-1]
    train_emb_all = emb_tr.reshape(-1, d)

    l2_maps, _ = search_nn(emb_te, train_emb_all, method=method, nn=nn)
    anomaly_maps = np.mean(l2_maps, axis=-1)

    return anomaly_maps



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Aggregator module."""
import queue
import time
from logging import getLogger

from openfl.component.straggler_handling_functions import CutoffTimeBasedStragglerHandling
from openfl.databases import TensorDB
from openfl.interface.aggregation_functions import WeightedAverage
from openfl.pipelines import NoCompressionPipeline, GenericPipeline
from openfl.pipelines import TensorCodec
from openfl.protocols import base_pb2
from openfl.protocols import utils
from openfl.utilities import TaskResultKey
from openfl.utilities import TensorKey
from openfl.utilities import change_tags
from openfl.utilities.logs import write_metric


class Aggregator:
    r"""An Aggregator is the central node in federated learning.

    Args:
        aggregator_uuid (str): Aggregation ID.
        federation_uuid (str): Federation ID.
        authorized_cols (list of str): The list of IDs of enrolled collaborators.
        init_state_path* (str): The location of the initial weight file.
        last_state_path* (str): The file location to store the latest weight.
        best_state_path* (str): The file location to store the weight of the best model.
        db_store_rounds* (int): Rounds to store in TensorDB.

    Note:
        \* - plan setting.
    """

    def __init__(self,

                 aggregator_uuid,
                 federation_uuid,
                 authorized_cols,

                 init_state_path,
                 best_state_path,
                 last_state_path,

                 assigner,
                 straggler_handling_policy=None,
                 rounds_to_train=256,
                 single_col_cert_common_name=None,
                 compression_pipeline=None,
                 db_store_rounds=1,
                 write_logs=False,
                 nn=True,
                 **kwargs):
        """Initialize."""
        self.round_number = 0
        self.single_col_cert_common_name = single_col_cert_common_name

        self.nn = nn

        if self.single_col_cert_common_name is not None:
            self._log_big_warning()
        else:
            # FIXME: '' instead of None is just for protobuf compatibility.
            # Cleaner solution?
            self.single_col_cert_common_name = ''

        self.straggler_handling_policy = straggler_handling_policy or CutoffTimeBasedStragglerHandling()
        self._end_of_round_check_done = [False] * rounds_to_train
        self.stragglers_for_task = {}

        self.rounds_to_train = rounds_to_train

        # if the collaborator requests a delta, this value is set to true
        self.authorized_cols = authorized_cols
        self.uuid = aggregator_uuid
        self.federation_uuid = federation_uuid
        self.assigner = assigner
        self.quit_job_sent_to = []

        self.tensor_db = TensorDB(self.nn, agg=True)
        # FIXME: I think next line generates an error on the second round
        # if it is set to 1 for the aggregator.
        self.db_store_rounds = db_store_rounds

        # Gathered together logging-related objects
        self.write_logs = write_logs
        if self.write_logs:
            self.log_metric = write_metric
        self.logger = getLogger(__name__)
        self.best_model_score = None
        self.metric_queue = queue.Queue()

        if self.nn:
            self.compression_pipeline = compression_pipeline or NoCompressionPipeline()
        else:
            self.compression_pipeline = GenericPipeline(self.nn)

        self.tensor_codec = TensorCodec(self.compression_pipeline)

        self.init_state_path = init_state_path
        self.best_state_path = best_state_path
        self.last_state_path = last_state_path

        self.best_tensor_dict: dict = {}
        self.last_tensor_dict: dict = {}

        if kwargs.get('initial_tensor_dict', None) is not None:
            self._load_initial_tensors_from_dict(kwargs['initial_tensor_dict'])
            self.model = utils.construct_model_proto(
                tensor_dict=kwargs['initial_tensor_dict'],
                round_number=0,
                tensor_pipe=self.compression_pipeline)
        else:
            self.model: base_pb2.ModelProto = utils.load_proto(self.init_state_path)
            self._load_initial_tensors()  # keys are TensorKeys

        self.collaborator_tensor_results = {}  # {TensorKey: nparray}}

        # these enable getting all tensors for a task
        self.collaborator_tasks_results = {}  # {TaskResultKey: list of TensorKeys}

        self.collaborator_task_weight = {}  # {TaskResultKey: data_size}

    def _load_initial_tensors(self):
        """
        Load all of the tensors required to begin federated learning.

        Required tensors are: \
            1. Initial model.

        Returns:
            None
        """
        tensor_dict, round_number = utils.deconstruct_model_proto(
            self.model, compression_pipeline=self.compression_pipeline)

        if round_number > self.round_number:
            self.logger.info(
                f'Starting training from round {round_number} of previously saved model'
            )
            self.round_number = round_number
        tensor_key_dict = {
            TensorKey(k, self.uuid, self.round_number, False, ('model',)):
                v for k, v in tensor_dict.items()
        }
        # all initial model tensors are loaded here
        self.tensor_db.cache_tensor(tensor_key_dict)
        self.logger.debug(f'This is the initial tensor_db: {self.tensor_db}')

    def _load_initial_tensors_from_dict(self, tensor_dict):
        """
        Load all of the tensors required to begin federated learning.

        Required tensors are: \
            1. Initial model.

        Returns:
            None
        """
        # This could be changed to the same name for both cases
        key = ('model',) if self.nn else ('weak_learner',)
        tensor_key_dict = {
            TensorKey(k, self.uuid, self.round_number, False, key):
                v for k, v in tensor_dict.items()
        }
        # all initial model tensors are loaded here
        self.tensor_db.cache_tensor(tensor_key_dict)
        self.logger.debug(f'This is the initial tensor_db: {self.tensor_db}')

    def _save_model(self, round_number, file_path):
        """
        Save the best or latest model.

        Args:
            round_number: int
                Model round to be saved
            file_path: str
                Either the best model or latest model file path

        Returns:
            None
        """
        # Extract the model from TensorDB and set it to the new model
        og_tensor_dict, _ = utils.deconstruct_model_proto(
            self.model, compression_pipeline=self.compression_pipeline)
        tensor_keys = [
            TensorKey(
                k, self.uuid, round_number, False, ('model',)
            ) for k, v in og_tensor_dict.items()
        ]
        tensor_dict = {}
        for tk in tensor_keys:
            tk_name, _, _, _, _ = tk
            tensor_dict[tk_name] = self.tensor_db.get_tensor_from_cache(tk)
            if tensor_dict[tk_name] is None:
                self.logger.info(f'Cannot save model for round {round_number}. Continuing...')
                return
        if file_path == self.best_state_path:
            self.best_tensor_dict = tensor_dict
        if file_path == self.last_state_path:
            self.last_tensor_dict = tensor_dict
        self.model = utils.construct_model_proto(
            tensor_dict, round_number, self.compression_pipeline)
        utils.dump_proto(self.model, file_path)

    def valid_collaborator_cn_and_id(self, cert_common_name,
                                     collaborator_common_name):
        """
        Determine if the collaborator certificate and ID are valid for this federation.

        Args:
            cert_common_name: Common name for security certificate
            collaborator_common_name: Common name for collaborator

        Returns:
            bool: True means the collaborator common name matches the name in
                  the security certificate.

        """
        # if self.test_mode_whitelist is None, then the common_name must
        # match collaborator_common_name and be in authorized_cols
        # FIXME: '' instead of None is just for protobuf compatibility.
        #  Cleaner solution?
        if self.single_col_cert_common_name == '':
            return (cert_common_name == collaborator_common_name
                    and collaborator_common_name in self.authorized_cols)
        # otherwise, common_name must be in whitelist and
        # collaborator_common_name must be in authorized_cols
        else:
            return (cert_common_name == self.single_col_cert_common_name
                    and collaborator_common_name in self.authorized_cols)

    def all_quit_jobs_sent(self):
        """Assert all quit jobs are sent to collaborators."""
        return set(self.quit_job_sent_to) == set(self.authorized_cols)

    @staticmethod
    def _get_sleep_time():
        """
        Sleep 10 seconds.

        Returns:
            sleep_time: int
        """
        # Decrease sleep period for finer discretezation
        return 1

    def _time_to_quit(self):
        """
        If all rounds are complete, it's time to quit.

        Returns:
            is_time_to_quit: bool
        """
        if self.round_number >= self.rounds_to_train:
            return True
        return False

    def get_tasks(self, collaborator_name):
        """
        RPC called by a collaborator to determine which tasks to perform.

        Args:
            collaborator_name: str
                Requested collaborator name

        Returns:
            tasks: list[str]
                List of tasks to be performed by the requesting collaborator
                for the current round.
            sleep_time: int
            time_to_quit: bool
        """
        self.logger.debug(
            f'Aggregator GetTasks function reached from collaborator {collaborator_name}...'
        )

        # first, if it is time to quit, inform the collaborator
        if self._time_to_quit():
            self.logger.info(f'Sending signal to collaborator {collaborator_name} to shutdown...')
            self.quit_job_sent_to.append(collaborator_name)

            tasks = None
            sleep_time = 0
            time_to_quit = True

            return tasks, self.round_number, sleep_time, time_to_quit

        time_to_quit = False
        # otherwise, get the tasks from our task assigner
        tasks = self.assigner.get_tasks_for_collaborator(collaborator_name, self.round_number)

        # if no tasks, tell the collaborator to sleep
        if len(tasks) == 0:
            tasks = None
            sleep_time = self._get_sleep_time()

            return tasks, self.round_number, sleep_time, time_to_quit

        # if we do have tasks, remove any that we already have results for
        if isinstance(tasks[0], str):
            # backward compatibility
            tasks = [
                t for t in tasks if not self._collaborator_task_completed(
                    collaborator_name, t, self.round_number)
            ]
            for t in tasks:
                if t in self.stragglers_for_task:
                    for stragglers in self.stragglers_for_task[t]:
                        if collaborator_name in stragglers:
                            tasks.remove(t)
        else:
            tasks = [
                t for t in tasks if not self._collaborator_task_completed(
                    collaborator_name, t.name, self.round_number)
            ]
            for t in tasks:
                if t.name in self.stragglers_for_task:
                    for stragglers in self.stragglers_for_task[t.name]:
                        if collaborator_name in stragglers:
                            tasks.remove(t)

        # Do the check again because it's possible that all tasks have
        # been completed
        if len(tasks) == 0:
            tasks = None
            sleep_time = self._get_sleep_time()

            return tasks, self.round_number, sleep_time, time_to_quit

        self.logger.info(
            f'Sending tasks to collaborator {collaborator_name} for round {self.round_number}'
        )
        sleep_time = 0

        if hasattr(self.straggler_handling_policy, 'round_start_time'):
            self.straggler_handling_policy.round_start_time = time.time()

        return tasks, self.round_number, sleep_time, time_to_quit

    def get_tensor(self, collaborator_name, tensor_name,
                   round_number, report, tags, require_lossless, aggregated=True):
        """
        RPC called by collaborator.

        Performs local lookup to determine if there is an aggregated tensor available \
            that matches the request.

        Args:
            collaborator_name : str
                Requested tensor key collaborator name
            tensor_name: str
            require_lossless: bool
            round_number: int
            report: bool
            tags: tuple[str, ...]
        Returns:
            named_tensor : protobuf NamedTensor
                the tensor requested by the collaborator
        """
        self.logger.debug(f'Retrieving tensor {tensor_name},{round_number},{tags} '
                          f'for collaborator {collaborator_name}')

        if 'compressed' in tags or require_lossless:
            compress_lossless = True
        else:
            compress_lossless = False

        # TODO the TensorDB doesn't support compressed data yet.
        #  The returned tensor will
        # be recompressed anyway.
        if 'compressed' in tags:
            tags = change_tags(tags, remove_field='compressed')
        if 'lossy_compressed' in tags:
            tags = change_tags(tags, remove_field='lossy_compressed')

        if aggregated:
            self.logger.debug(f'Retrieving aggregated tensor {tensor_name},{round_number},{tags} '
                              f'for collaborator {collaborator_name}')
            tensor_key = TensorKey(
                tensor_name, self.uuid, round_number, report, tags
            )
        else:
            self.logger.debug(f'Retrieving tensor {tensor_name},{round_number},{tags} '
                              f'for collaborator {collaborator_name}')
            # TODO: this should be more clear
            # AdaBoost.F TensorKey handling
            tensor_key = TensorKey(
                'errors' if 'adaboost_coeff' in tags else tensor_name, self.uuid,
                0 if 'weak_learner' in tags else round_number, True if 'adaboost_coeff' in tags else report, tuple(tags)
            )
        tensor_name, origin, round_number, report, tags = tensor_key

        if 'aggregated' in tags and 'delta' in tags and round_number != 0:
            agg_tensor_key = TensorKey(
                tensor_name, origin, round_number, report, ('aggregated',)
            )
        else:
            agg_tensor_key = tensor_key

        nparray = self.tensor_db.get_tensor_from_cache(agg_tensor_key)

        start_retrieving_time = time.time()
        while (nparray is None):
            self.logger.debug(f'Waiting for tensor_key {agg_tensor_key}')
            self.logger.debug(f'This is tensor_db: {self.tensor_db}')
            time.sleep(1)
            nparray = self.tensor_db.get_tensor_from_cache(agg_tensor_key)
            if (time.time() - start_retrieving_time) > 180:
                break

        if nparray is None:
            raise ValueError(f'Aggregator does not have an aggregated tensor for {tensor_key}')

        # quite a bit happens in here, including compression, delta handling,
        # etc...
        # we might want to cache these as well
        named_tensor = self._nparray_to_named_tensor(
            agg_tensor_key,
            nparray,
            send_model_deltas=True,
            compress_lossless=compress_lossless
        )

        return named_tensor

    def _nparray_to_named_tensor(self, tensor_key, nparray, send_model_deltas,
                                 compress_lossless):
        """
        Construct the NamedTensor Protobuf.

        Also includes logic to create delta, compress tensors with the TensorCodec, etc.
        """
        tensor_name, origin, round_number, report, tags = tensor_key
        # if we have an aggregated tensor, we can make a delta
        if 'aggregated' in tags and send_model_deltas:
            # Should get the pretrained model to create the delta. If training
            # has happened, Model should already be stored in the TensorDB
            model_tk = TensorKey(tensor_name,
                                 origin,
                                 round_number - 1,
                                 report,
                                 ('model',))

            model_nparray = self.tensor_db.get_tensor_from_cache(model_tk)

            assert (model_nparray is not None), (
                'The original model layer should be present if the latest '
                'aggregated model is present')
            delta_tensor_key, delta_nparray = self.tensor_codec.generate_delta(
                tensor_key,
                nparray,
                model_nparray
            )
            delta_comp_tensor_key, delta_comp_nparray, metadata = self.tensor_codec.compress(
                delta_tensor_key,
                delta_nparray,
                lossless=compress_lossless
            )
            named_tensor = utils.construct_named_tensor(
                delta_comp_tensor_key,
                delta_comp_nparray,
                metadata,
                lossless=compress_lossless
            )

        else:
            # Assume every other tensor requires lossless compression
            compressed_tensor_key, compressed_nparray, metadata = self.tensor_codec.compress(
                tensor_key,
                nparray,
                require_lossless=True,
            )
            named_tensor = utils.construct_named_tensor(
                compressed_tensor_key,
                compressed_nparray,
                metadata,
                lossless=compress_lossless
            )

        return named_tensor

    def _collaborator_task_completed(self, collaborator, task_name, round_num):
        """
        Check if the collaborator has completed the task for the round.

        The aggregator doesn't actually know which tensors should be sent from the collaborator \
            so it must to rely specifically on the presence of previous results

        Args:
            collaborator : str
                collaborator to check if their task has been completed
            task_name : str
                The name of the task (TaskRunner function)
            round_num : int

        Returns:
            task_competed : bool
                Whether or not the collaborator has completed the task for this
                round
        """
        task_key = TaskResultKey(task_name, collaborator, round_num)
        return task_key in self.collaborator_tasks_results

    def send_local_task_results(self, collaborator_name, round_number, task_name,
                                data_size, named_tensors):
        """
        RPC called by collaborator.

        Transmits collaborator's task results to the aggregator.

        Args:
            collaborator_name: str
            task_name: str
            round_number: int
            data_size: int
            named_tensors: protobuf NamedTensor
        Returns:
             None
        """
        if self._time_to_quit() or self._is_task_done(task_name):
            self.logger.warning(
                f'STRAGGLER: Collaborator {collaborator_name} is reporting results after task {task_name} has finished.'
            )
            return

        if self.round_number != round_number:
            self.logger.warning(
                f'Collaborator {collaborator_name} is reporting results for the wrong round: {round_number}.'
                f' Ignoring...'
            )
            return

        self.logger.info(
            f'Collaborator {collaborator_name} is sending task results '
            f'for {task_name}, round {round_number}'
        )

        task_key = TaskResultKey(task_name, collaborator_name, round_number)

        # we mustn't have results already
        if self._collaborator_task_completed(
                collaborator_name, task_name, round_number
        ):
            raise ValueError(
                f'Aggregator already has task results from collaborator {collaborator_name}'
                f' for task {task_key}'
            )

        # initialize the list of tensors that go with this task
        # Setting these incrementally is leading to missing values
        task_results = []

        # go through the tensors and add them to the tensor dictionary and the
        # task dictionary
        for named_tensor in named_tensors:
            # quite a bit happens in here, including decompression, delta
            # handling, etc...
            tensor_key, nparray = self._process_named_tensor(
                named_tensor, collaborator_name
            )
            if 'metric' in tensor_key.tags:
                if self.nn:
                    nparray = nparray.item()
                metric_dict = {
                    'metric_origin': tensor_key.tags[-1],
                    'task_name': task_name,
                    'metric_name': tensor_key.tensor_name,
                    'metric_value': nparray,
                    'round': round_number}
                if self.write_logs:
                    self.log_metric(tensor_key.tags[-1], task_name,
                                    tensor_key.tensor_name, nparray, round_number)

                self.logger.metric(f'Round {round_number}, '
                                   f'collaborator {tensor_key.tags[-1]} '
                                   f'{task_name} result '
                                   f'{tensor_key.tensor_name}:\t{nparray}')
                # TODO: this should be reintroduced
                if self.nn:
                    self.metric_queue.put(metric_dict)

            task_results.append(tensor_key)
            # By giving task_key it's own weight, we can support different
            # training/validation weights
            # As well as eventually supporting weights that change by round
            # (if more data is added)
            self.collaborator_task_weight[task_key] = data_size

        self.collaborator_tasks_results[task_key] = task_results

        self._end_of_task_check(task_name)

    def synch(self, task_name, round_number):
        result = self._is_task_done(task_name, round_number)
        return result

    def _process_named_tensor(self, named_tensor, collaborator_name):
        """
        Extract the named tensor fields.

        Performs decompression, delta computation, and inserts results into TensorDB.

        Args:
            named_tensor:       NamedTensor (protobuf)
                protobuf that will be extracted from and processed
            collaborator_name:  str
                Collaborator name is needed for proper tagging of resulting
                tensorkeys

        Returns:
            tensor_key : TensorKey (named_tuple)
                The tensorkey extracted from the protobuf
            nparray : np.array
                The numpy array associated with the returned tensorkey
        """
        raw_bytes = named_tensor.data_bytes
        if self.nn:
            metadata = [{'int_to_float': proto.int_to_float,
                         'int_list': proto.int_list,
                         'bool_list': proto.bool_list}
                        for proto in named_tensor.transformer_metadata]
        else:
            metadata = [{'int_to_float': proto.int_to_float,
                         'int_list': proto.int_list,
                         'bool_list': proto.bool_list,
                         'model': proto.model}
                        for proto in named_tensor.transformer_metadata]

        # The tensor has already been transfered to aggregator,
        # so the newly constructed tensor should have the aggregator origin
        tensor_key = TensorKey(
            named_tensor.name,
            self.uuid,
            named_tensor.round_number,
            named_tensor.report,
            tuple(named_tensor.tags)
        )
        tensor_name, origin, round_number, report, tags = tensor_key
        assert ('compressed' in tags or 'lossy_compressed' in tags), (
            f'Named tensor {tensor_key} is not compressed'
        )
        if 'compressed' in tags:
            dec_tk, decompressed_nparray = self.tensor_codec.decompress(
                tensor_key,
                data=raw_bytes,
                transformer_metadata=metadata,
                require_lossless=True
            )
            dec_name, dec_origin, dec_round_num, dec_report, dec_tags = dec_tk
            # Need to add the collaborator tag to the resulting tensor
            new_tags = change_tags(dec_tags, add_field=collaborator_name)

            # layer.agg.n.trained.delta.col_i
            decompressed_tensor_key = TensorKey(
                dec_name, dec_origin, dec_round_num, dec_report, new_tags
            )
        if 'lossy_compressed' in tags:
            dec_tk, decompressed_nparray = self.tensor_codec.decompress(
                tensor_key,
                data=raw_bytes,
                transformer_metadata=metadata,
                require_lossless=False
            )
            dec_name, dec_origin, dec_round_num, dec_report, dec_tags = dec_tk
            new_tags = change_tags(dec_tags, add_field=collaborator_name)
            # layer.agg.n.trained.delta.lossy_decompressed.col_i
            decompressed_tensor_key = TensorKey(
                dec_name, dec_origin, dec_round_num, dec_report, new_tags
            )

        if 'delta' in tags:
            base_model_tensor_key = TensorKey(
                tensor_name, origin, round_number, report, ('model',)
            )
            base_model_nparray = self.tensor_db.get_tensor_from_cache(
                base_model_tensor_key
            )
            if base_model_nparray is None:
                raise ValueError(f'Base model {base_model_tensor_key} not present in TensorDB')
            final_tensor_key, final_nparray = self.tensor_codec.apply_delta(
                decompressed_tensor_key,
                decompressed_nparray, base_model_nparray
            )
        else:
            final_tensor_key = decompressed_tensor_key
            final_nparray = decompressed_nparray

        assert (final_nparray is not None), f'Could not create tensorkey {final_tensor_key}'
        self.tensor_db.cache_tensor({final_tensor_key: final_nparray})
        self.logger.debug(f'Created TensorKey: {final_tensor_key}')

        return final_tensor_key, final_nparray

    def _end_of_task_check(self, task_name):
        """
        Check whether all collaborators who are supposed to perform the task complete.

        Args:
            task_name : str
                The task name to check

        Returns:
            complete : boolean
                Is the task done
        """
        if self._is_task_done(task_name):
            # now check for the end of the round
            self._end_of_round_check(task_name)

    def _prepare_trained(self, tensor_name, origin, round_number, report, agg_results):
        """
        Prepare aggregated tensorkey tags.

        Args:
           tensor_name : str
           origin:
           round_number: int
           report: bool
           agg_results: np.array
        """
        # The aggregated tensorkey tags should have the form of
        # 'trained' or 'trained.lossy_decompressed'
        # They need to be relabeled to 'aggregated' and
        # reinserted. Then delta performed, compressed, etc.
        # then reinserted to TensorDB with 'model' tag

        # First insert the aggregated model layer with the
        # correct tensorkey
        agg_tag_tk = TensorKey(
            tensor_name,
            origin,
            round_number + 1 if self.nn else round_number,
            report,
            ('aggregated',)
        )
        self.tensor_db.cache_tensor({agg_tag_tk: agg_results})

        # Create delta and save it in TensorDB
        base_model_tk = TensorKey(
            tensor_name,
            origin,
            round_number,
            report,
            ('model',)
        )
        base_model_nparray = self.tensor_db.get_tensor_from_cache(base_model_tk)
        if self.nn and base_model_nparray is not None:
            delta_tk, delta_nparray = self.tensor_codec.generate_delta(
                agg_tag_tk,
                agg_results,
                base_model_nparray
            )
        else:
            # This condition is possible for base model
            # optimizer states (i.e. Adam/iter:0, SGD, etc.)
            # These values couldn't be present for the base
            # model because no training occurs on the aggregator
            delta_tk, delta_nparray = agg_tag_tk, agg_results

        # Compress lossless/lossy
        compressed_delta_tk, compressed_delta_nparray, metadata = self.tensor_codec.compress(
            delta_tk, delta_nparray
        )

        # TODO extend the TensorDB so that compressed data is
        #  supported. Once that is in place
        # the compressed delta can just be stored here instead
        # of recreating it for every request

        # Decompress lossless/lossy
        decompressed_delta_tk, decompressed_delta_nparray = self.tensor_codec.decompress(
            compressed_delta_tk,
            compressed_delta_nparray,
            metadata
        )

        self.tensor_db.cache_tensor({decompressed_delta_tk: decompressed_delta_nparray})

        # Apply delta (unless delta couldn't be created)
        if self.nn and base_model_nparray is not None:
            self.logger.debug(f'Applying delta for layer {decompressed_delta_tk[0]}')
            new_model_tk, new_model_nparray = self.tensor_codec.apply_delta(
                decompressed_delta_tk,
                decompressed_delta_nparray,
                base_model_nparray
            )
        else:
            new_model_tk, new_model_nparray = decompressed_delta_tk, decompressed_delta_nparray

        # Now that the model has been compressed/decompressed
        # with delta operations,
        # Relabel the tags to 'model'
        (new_model_tensor_name, new_model_origin, new_model_round_number,
         new_model_report, new_model_tags) = new_model_tk
        final_model_tk = TensorKey(
            new_model_tensor_name,
            new_model_origin,
            new_model_round_number,
            new_model_report,
            ('model',)
        )

        # Finally, cache the updated model tensor
        self.tensor_db.cache_tensor({final_model_tk: new_model_nparray})

    def _compute_validation_related_task_metrics(self, task_name):
        """
        Compute all validation related metrics.

        Args:
            task_name : str
                The task name to compute
        """
        # By default, print out all of the metrics that the validation
        # task sent
        # This handles getting the subset of collaborators that may be
        # part of the validation task
        all_collaborators_for_task = self.assigner.get_collaborators_for_task(
            task_name, self.round_number
        )
        # leave out stragglers for the round
        collaborators_for_task = []
        for c in all_collaborators_for_task:
            if self._collaborator_task_completed(c, task_name, self.round_number):
                collaborators_for_task.append(c)

        # The collaborator data sizes for that task
        collaborator_weights_unnormalized = {
            c: self.collaborator_task_weight[TaskResultKey(task_name, c, self.round_number)]
            for c in collaborators_for_task}
        weight_total = sum(collaborator_weights_unnormalized.values())
        collaborator_weight_dict = {
            k: v / weight_total
            for k, v in collaborator_weights_unnormalized.items()
        }

        # The validation task should have just a couple tensors (i.e.
        # metrics) associated with it. Because each collaborator should
        # have sent the same tensor list, we can use the first
        # collaborator in our subset, and apply the correct
        # transformations to the tensorkey to resolve the aggregated
        # tensor for that round
        task_agg_function = self.assigner.get_aggregation_type_for_task(task_name)
        task_key = TaskResultKey(task_name, collaborators_for_task[0], self.round_number)
        for tensor_key in self.collaborator_tasks_results[task_key]:
            tensor_name, origin, round_number, report, tags = tensor_key
            assert (collaborators_for_task[0] in tags), (
                f'Tensor {tensor_key} in task {task_name} has not been processed correctly'
            )
            # Strip the collaborator label, and lookup aggregated tensor
            new_tags = change_tags(tags, remove_field=collaborators_for_task[0])
            agg_tensor_key = TensorKey(tensor_name, origin, round_number, report, new_tags)
            agg_tensor_name, agg_origin, agg_round_number, agg_report, agg_tags = agg_tensor_key
            # TODO: This if can be removed (maybe with a well-configurated plan)
            if 'metric' in tags:
                if task_name == "2_weak_learners_validate":
                    agg_function = task_agg_function
                else:
                    agg_function = WeightedAverage()
            else:
                agg_function = task_agg_function
            agg_results = self.tensor_db.get_aggregated_tensor(
                agg_tensor_key, collaborator_weight_dict, aggregation_function=agg_function)
            if report:
                # Print the aggregated metric
                metric_dict = {
                    'metric_origin': 'Aggregator',
                    'task_name': task_name,
                    'metric_name': tensor_key.tensor_name,
                    'metric_value': agg_results.item() if self.nn else agg_results,
                    'round': round_number}

                if agg_results is None:
                    self.logger.warning(
                        f'Aggregated metric {agg_tensor_name} could not be collected '
                        f'for round {self.round_number}. Skipping reporting for this round')
                if agg_function:
                    self.logger.metric(f'Round {round_number}, aggregator: {task_name} '
                                       f'{agg_function} {agg_tensor_name}:\t{agg_results}')
                else:
                    self.logger.metric(f'Round {round_number}, aggregator: {task_name} '
                                       f'{agg_tensor_name}:\t{agg_results}')
                if self.write_logs:
                    self.log_metric('Aggregator', task_name,
                                    tensor_key.tensor_name,
                                    agg_results, round_number)
                # TODO: this should be reintroduced
                if self.nn:
                    self.metric_queue.put(metric_dict)
                # TODO: This is disabled for non DNN models since it is in conflict with the exchange of arrays of metrics
                if self.nn:
                    # TODO Add all of the logic for saving the model based
                    #  on best accuracy, lowest loss, etc.
                    if 'validate_agg' in tags:
                        # Compare the accuracy of the model, and
                        # potentially save it
                        if self.best_model_score is None or self.best_model_score < agg_results:
                            self.logger.metric(f'Round {round_number}: saved the best '
                                               f'model with score {agg_results:f}')
                            self.best_model_score = agg_results
                            self._save_model(round_number, self.best_state_path)
            if 'trained' in tags:
                self._prepare_trained(tensor_name, origin, round_number, report, agg_results)
            # TODO: generalize this possibly
            if task_name == '2_weak_learners_validate':
                self._prepare_adaboost(tensor_name, origin, round_number, report, agg_results)

    def _end_of_round_check(self, task_name=None):
        """
        Check if the round complete.

        If so, perform many end of round operations,
        such as model aggregation, metric reporting, delta generation (+
        associated tensorkey labeling), and save the model

        Args:
            None

        Returns:
            None
        """
        if self.nn:
            if not self._is_round_done() or self._end_of_round_check_done[self.round_number]:
                return
            all_tasks = self.assigner.get_all_tasks_for_round(self.round_number)
            for task_name in all_tasks:
                self._compute_validation_related_task_metrics(task_name)
        else:
            self._compute_validation_related_task_metrics(task_name)
            if not self._is_round_done():
                return

        # Once all of the task results have been processed
        self._end_of_round_check_done[self.round_number] = True
        self.round_number += 1
        # resetting stragglers for task for a new round
        self.stragglers_for_task = {}

        # Save the latest model
        if self.nn:
            self.logger.info(f'Saving round {self.round_number} model...')
            self._save_model(self.round_number, self.last_state_path)

        # TODO This needs to be fixed!
        if self._time_to_quit():
            self.logger.info('Experiment Completed. Cleaning up...')
        else:
            self.logger.info(f'Starting round {self.round_number}...')

        # Cleaning tensor db
        self.tensor_db.clean_up(self.db_store_rounds)

    def _is_task_done(self, task_name, round_number=None):
        """Check that task is done."""

        if round_number is None:
            round_number = self.round_number

        all_collaborators = self.assigner.get_collaborators_for_task(
            task_name, round_number
        )

        collaborators_done = []
        for c in all_collaborators:
            if self._collaborator_task_completed(
                    c, task_name, round_number):
                collaborators_done.append(c)

        straggler_check = self.straggler_handling_policy.straggler_cutoff_check(
            len(collaborators_done), all_collaborators)

        stragglers = []
        if straggler_check:
            for c in all_collaborators:
                if c not in collaborators_done:
                    stragglers.append(c)
            self.logger.info('\tEnding task {} early due to straggler cutoff policy'.format(task_name))
            self.logger.warning('\tIdentified stragglers: {} for task {}'.format(stragglers, task_name))
            self.stragglers_for_task[task_name] = stragglers

        # all are done or straggler policy calls for early round end.
        return straggler_check or len(all_collaborators) == len(collaborators_done)

    def _is_round_done(self):
        """Check that round is done."""
        tasks_for_round = self.assigner.get_all_tasks_for_round(self.round_number)

        return all([
            self._is_task_done(
                task_name) for task_name in tasks_for_round])

    def _log_big_warning(self):
        """Warn user about single collaborator cert mode."""
        self.logger.warning(
            f'\n{the_dragon}\nYOU ARE RUNNING IN SINGLE COLLABORATOR CERT MODE! THIS IS'
            f' NOT PROPER PKI AND '
            f'SHOULD ONLY BE USED IN DEVELOPMENT SETTINGS!!!! YE HAVE BEEN'
            f' WARNED!!!'
        )

    def _prepare_adaboost(self, tensor_name, origin, round_number, report, agg_results):
        """
        Prepare aggregated tensorkey tags for adaboost.
        Args:
           tensor_name : str
           origin:
           round_number: int
           report: bool
           agg_results: np.array
        """
        # The aggregated tensorkey tags should have the form of
        # 'trained' or 'trained.lossy_decompressed'
        # They need to be relabeled to 'aggregated' and
        # reinserted. Then delta performed, compressed, etc.
        # then reinserted to TensorDB with 'model' tag

        # First insert the aggregated model layer with the
        # correct tensorkey
        agg_tag_tk = TensorKey(
            tensor_name,
            origin,
            round_number,
            report,
            ('adaboost_coeff',)
        )

        self.tensor_db.cache_tensor({agg_tag_tk: agg_results})

    def stop(self, failed_collaborator: str = None) -> None:
        """Stop aggregator execution."""
        self.logger.info('Force stopping the aggregator execution.')
        for collaborator_name in filter(lambda c: c != failed_collaborator, self.authorized_cols):
            self.logger.info(f'Sending signal to collaborator {collaborator_name} to shutdown...')
            self.quit_job_sent_to.append(collaborator_name)


the_dragon = '''

 ,@@.@@+@@##@,@@@@.`@@#@+  *@@@@ #@##@  `@@#@# @@@@@   @@    @@@@` #@@@ :@@ `@#`@@@#.@
  @@ #@ ,@ +. @@.@* #@ :`   @+*@ .@`+.   @@ *@::@`@@   @@#  @@  #`;@`.@@ @@@`@`#@* +:@`
  @@@@@ ,@@@  @@@@  +@@+    @@@@ .@@@    @@ .@+:@@@:  .;+@` @@ ,;,#@` @@ @@@@@ ,@@@* @
  @@ #@ ,@`*. @@.@@ #@ ,;  `@+,@#.@.*`   @@ ,@::@`@@` @@@@# @@`:@;*@+ @@ @`:@@`@ *@@ `
 .@@`@@,+@+;@.@@ @@`@@;*@  ;@@#@:*@+;@  `@@;@@ #@**@+;@ `@@:`@@@@  @@@@.`@+ .@ +@+@*,@
  `` ``     ` ``  .     `     `      `     `    `  .` `  ``   ``    ``   `       .   `



                                            .**
                                      ;`  `****:
                                     @**`*******
                         ***        +***********;
                        ,@***;` .*:,;************
                        ;***********@@***********
                        ;************************,
                        `*************************
                         *************************
                         ,************************
                          **#*********************
                          *@****`     :**********;
                          +**;          .********.
                          ;*;            `*******#:                       `,:
                                          ****@@@++::                ,,;***.
                                          *@@@**;#;:         +:      **++*,
                                          @***#@@@:          +*;     ,****
                                          @*@+****           ***`     ****,
                                         ,@#******.  ,       ****     **;,**.
                                         * ******** :,       ;*:*+    **  :,**
                                        #  ********::      *,.*:**`   *      ,*;
                                        .  *********:      .+,*:;*:   :      `:**
                                       ;   :********:       ***::**   `       ` **
                                       +   :****::***  ,    *;;::**`             :*
                                      ``   .****::;**:::    *;::::*;              ;*
                                      *     *****::***:.    **::::**               ;:
                                      #     *****;:****     ;*::;***               ,*`
                                      ;     ************`  ,**:****;               ::*
                                      :     *************;:;*;*++:                   *.
                                      :     *****************;*                      `*
                                     `.    `*****************;  :                     *.
                                     .`    .*+************+****;:                     :*
                                     `.    :;+***********+******;`    :              .,*
                                      ;    ::*+*******************. `::              .`:.
                                      +    :::**********************;;:`                *
                                      +    ,::;*************;:::*******.                *
                                      #    `:::+*************:::;********  :,           *
                                      @     :::***************;:;*********;:,           *
                                      @     ::::******:*********************:         ,:*
                                      @     .:::******:;*********************,         :*
                                      #      :::******::******###@*******;;****        *,
                                      #      .::;*****::*****#****@*****;:::***;  ``  **
                                      *       ::;***********+*****+#******::*****,,,,**
                                      :        :;***********#******#******************
                                      .`       `;***********#******+****+************
                                      `,        ***#**@**+***+*****+**************;`
                                       ;         *++**#******#+****+`      `.,..
                                       +         `@***#*******#****#
                                       +          +***@********+**+:
                                       *         .+**+;**;;;**;#**#
                                      ,`         ****@         +*+:
                                      #          +**+         :+**
                                      @         ;**+,       ,***+
                                      #      #@+****      *#****+
                                     `;     @+***+@      `#**+#++
                                     #      #*#@##,      .++:.,#
                                    `*      @#            +.
                                  @@@
                                 # `@
                                  ,                                                        '''



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Pytorch Framework Adapter plugin for multiple optimizers."""


from openfl.plugins.frameworks_adapters.pytorch_adapter import _get_optimizer_state
from openfl.plugins.frameworks_adapters.pytorch_adapter import FrameworkAdapterPlugin
from openfl.plugins.frameworks_adapters.pytorch_adapter import to_cpu_numpy


class FrameworkAdapterPluginforMultipleOpt(FrameworkAdapterPlugin):
    """Framework adapter plugin class for multiple optimizers."""

    def __init__(self):
        """Initialize framework adapter."""
        super().__init__()

    @staticmethod
    def get_tensor_dict(model, optimizers=None):
        """
        Extract tensor dict from a model and a list of optimizers.

        Returns:
        dict {weight name: numpy ndarray}
        """
        state = to_cpu_numpy(model.state_dict())
        if optimizers is not None:
            for opt in optimizers:
                if isinstance(opt, dict):
                    opt_state = _get_optimizer_state(opt['optimizer'])
                else:
                    opt_state = _get_optimizer_state(opt)

                state = {**state, **opt_state}

        return state



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Straggler Handling functions package."""

from .straggler_handling_function import StragglerHandlingFunction
from .cutoff_time_based_straggler_handling import CutoffTimeBasedStragglerHandling
from .percentage_based_straggler_handling import PercentageBasedStragglerHandling

__all__ = ['CutoffTimeBasedStragglerHandling',
           'PercentageBasedStragglerHandling',
           'StragglerHandlingFunction']



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""You may copy this file as the starting point of your own model."""

import fastestimator as fe
import tensorflow as tf
from fastestimator.architecture.tensorflow import LeNet
from fastestimator.op.tensorop import Average
from fastestimator.op.tensorop.gradient import FGSM
from fastestimator.op.tensorop.gradient import Watch
from fastestimator.op.tensorop.loss import CrossEntropy
from fastestimator.op.tensorop.model import ModelOp
from fastestimator.op.tensorop.model import UpdateOp
from fastestimator.trace.metric import Accuracy

from openfl.federated import FastEstimatorTaskRunner
from openfl.federated import TaskRunner


class FastEstimatorFGSM(FastEstimatorTaskRunner):
    """An FGSM example based on the LeNet model."""

    def __init__(self, **kwargs):
        """
        Initialize.

        Args:
            **kwargs: Additional parameters to pass to the function
        """
        TaskRunner.__init__(self, **kwargs)
        # Now the data pipeline will be initialized and the rest of the
        # network/estimator can be built
        self.model = self.build_model()
        self.network = self.build_network()
        estimator = self.build_estimator()
        super().__init__(estimator, **kwargs)

        self.initialize_tensorkeys_for_functions()

        self.logger.info(self.model.__repr__())

        if self.data_loader is not None:
            self.logger.info(f'Train Set Size : {self.get_train_data_size()}')
            self.logger.info(f'Valid Set Size : {self.get_valid_data_size()}')

    def build_model(self):
        """
        Define the FastEstimator model architecture.

        Args:
            None

        Returns:
            model: Union[tf.keras.sequential, nn.module]

        """
        def get_model():
            tf.keras.backend.clear_session()  # to avoid layer names mismatching
            return LeNet(input_shape=(32, 32, 3))
        model = fe.build(model_fn=get_model,
                         optimizer_fn='adam', model_name='adv_model')
        return model

    def build_network(self):
        """
        Define the FastEstimator network flow.

        Args:
            None

        Returns:
            network: KerasNetwork object
        """
        epsilon = 0.04

        network = fe.Network(ops=[
            Watch(inputs='x'),
            ModelOp(model=self.model, inputs='x', outputs='y_pred'),
            CrossEntropy(inputs=('y_pred', 'y'), outputs='base_ce'),
            FGSM(data='x', loss='base_ce', outputs='x_adverse', epsilon=epsilon),
            ModelOp(model=self.model, inputs='x_adverse', outputs='y_pred_adv'),
            CrossEntropy(inputs=('y_pred_adv', 'y'), outputs='adv_ce'),
            Average(inputs=('base_ce', 'adv_ce'), outputs='avg_ce'),
            UpdateOp(model=self.model, loss_name='avg_ce')
        ])

        return network

    def build_estimator(self):
        """
        Define the estimator to run the experiment.

        This will persist throughout the lifetime of the TaskRunner.

        Args:
            None

        Returns:
            estimator: Estimator object
        """
        max_train_steps_per_epoch = None
        max_eval_steps_per_epoch = None

        traces = [
            Accuracy(true_key='y', pred_key='y_pred', output_name='clean_accuracy'),
            Accuracy(true_key='y', pred_key='y_pred_adv', output_name='adversarial_accuracy'),
        ]
        estimator = fe.Estimator(pipeline=self.data_loader.pipeline,
                                 network=self.network,
                                 epochs=2,
                                 traces=traces,
                                 max_train_steps_per_epoch=max_train_steps_per_epoch,
                                 max_eval_steps_per_epoch=max_eval_steps_per_epoch,
                                 monitor_names=['base_ce', 'adv_ce'],
                                 log_steps=1000)

        return estimator



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""TensorCodec module."""

import numpy as np

from openfl.pipelines import NoCompressionPipeline
from openfl.utilities import change_tags
from openfl.utilities import TensorKey


class TensorCodec:
    """TensorCodec is responsible for the following.

    1. Tracking the compression/decompression related dependencies of a given tensor
    2. Acting as a TensorKey aware wrapper for the compression_pipeline functionality
    """

    def __init__(self, compression_pipeline):
        """Initialize."""
        self.compression_pipeline = compression_pipeline
        if self.compression_pipeline.is_lossy():
            self.lossless_pipeline = NoCompressionPipeline()
        else:
            self.lossless_pipeline = compression_pipeline

    def set_lossless_pipeline(self, lossless_pipeline):
        """Set lossless pipeline."""
        assert lossless_pipeline.is_lossy() is False, (
            'The provided pipeline is not lossless')
        self.lossless_pipeline = lossless_pipeline

    def compress(self, tensor_key, data, require_lossless=False, **kwargs):
        """
        Function-wrapper around the tensor_pipeline.forward function.

        It also keeps track of the tensorkeys associated with the compressed nparray

        Args:
            tensor_key:             TensorKey is provided to verify it should
                                    be compressed, and new TensorKeys returned
                                    will be derivatives of the existing
                                    tensor_name

            data:                   (uncompressed) numpy array associated with
                                    the tensor_key

            require_lossless:       boolean. Does tensor require
                                    compression

        Returns:
            compressed_tensor_key:  Tensorkey corresponding to the decompressed
                                    tensor

            compressed_nparray:     The compressed tensor

            metadata:               metadata associated with compressed tensor

        """
        if require_lossless:
            compressed_nparray, metadata = self.lossless_pipeline.forward(
                data, **kwargs)
        else:
            compressed_nparray, metadata = self.compression_pipeline.forward(
                data, **kwargs)
        # Define the compressed tensorkey that should be
        # returned ('trained.delta'->'trained.delta.lossy_compressed')
        tensor_name, origin, round_number, report, tags = tensor_key
        if not self.compression_pipeline.is_lossy() or require_lossless:
            new_tags = change_tags(tags, add_field='compressed')
        else:
            new_tags = change_tags(tags, add_field='lossy_compressed')
        compressed_tensor_key = TensorKey(
            tensor_name, origin, round_number, report, new_tags)
        return compressed_tensor_key, compressed_nparray, metadata

    def decompress(self, tensor_key, data, transformer_metadata,
                   require_lossless=False, **kwargs):
        """
        Function-wrapper around the tensor_pipeline.backward function.

        It also keeps track of the tensorkeys associated with the decompressed nparray

        Args:
            tensor_key:             TensorKey is provided to verify it should
                                    be decompressed, and new TensorKeys
                                    returned will be derivatives of the
                                    existing tensor_name

            data:                   (compressed) numpy array associated with
                                    the tensor_key

            transformer_metadata:   metadata associated with the compressed
                                    tensor

            require_lossless:       boolean, does data require lossless
                                    decompression

        Returns:
            decompressed_tensor_key:    Tensorkey corresponding to the
                                        decompressed tensor

            decompressed_nparray:       The decompressed tensor

        """
        tensor_name, origin, round_number, report, tags = tensor_key

        assert (len(transformer_metadata) > 0), (
            'metadata must be included for decompression')
        assert (('compressed' in tags) or ('lossy_compressed' in tags)), (
            'Cannot decompress an uncompressed tensor')
        if require_lossless:
            assert ('compressed' in tags), (
                'Cannot losslessly decompress lossy tensor')

        if require_lossless or 'compressed' in tags:
            decompressed_nparray = self.lossless_pipeline.backward(
                data, transformer_metadata, **kwargs)
        else:
            decompressed_nparray = self.compression_pipeline.backward(
                data, transformer_metadata, **kwargs)
        # Define the decompressed tensorkey that should be returned
        if 'lossy_compressed' in tags:
            new_tags = change_tags(
                tags, add_field='lossy_decompressed', remove_field='lossy_compressed')
            decompressed_tensor_key = TensorKey(
                tensor_name, origin, round_number, report, new_tags)
        elif 'compressed' in tags:
            # 'compressed' == lossless compression; no need for
            # compression related tag after decompression
            new_tags = change_tags(tags, remove_field='compressed')
            decompressed_tensor_key = TensorKey(
                tensor_name, origin, round_number, report, new_tags)
        else:
            raise NotImplementedError(
                'Decompression is only supported on compressed data')

        return decompressed_tensor_key, decompressed_nparray

    @staticmethod
    def generate_delta(tensor_key, nparray, base_model_nparray):
        """
        Create delta from the updated layer and base layer.

        Args:
            tensor_key:         This is the tensor_key associated with the
                                nparray.
                                Should have a tag of 'trained' or 'aggregated'

            nparray:            The nparray that corresponds to the tensorkey

            base_model_nparray: The base model tensor that will be subtracted
                                from the new weights

        Returns:
            delta_tensor_key:   Tensorkey that corresponds to the delta weight
                                array

            delta:              Difference between the provided tensors

        """
        tensor_name, origin, round_number, report, tags = tensor_key
        if not np.isscalar(nparray):
            assert nparray.shape == base_model_nparray.shape, (
                f'Shape of updated layer ({nparray.shape}) is not equal to base '
                f'layer shape of ({base_model_nparray.shape})'
            )
        assert 'model' not in tags, (
            'The tensorkey should be provided '
            'from the layer with new weights, not the base model')
        new_tags = change_tags(tags, add_field='delta')
        delta_tensor_key = TensorKey(
            tensor_name, origin, round_number, report, new_tags)
        return delta_tensor_key, nparray - base_model_nparray

    @staticmethod
    def apply_delta(tensor_key, delta, base_model_nparray, creates_model=False):
        """
        Add delta to the nparray.

        Args:
            tensor_key:             This is the tensor_key associated with the
                                    delta. Should have a tag of 'trained' or
                                    'aggregated'
            delta:                  Weight delta between the new model and
                                    old model
            base_model_nparray:     The nparray that corresponds to the prior
                                    weights
            creates_model:          If flag is set, the tensorkey returned
                                    will correspond to the aggregator model

        Returns:
            new_model_tensor_key:   Latest model layer tensorkey
            new_model_nparray:      Latest layer weights

        """
        tensor_name, origin, round_number, report, tags = tensor_key
        if not np.isscalar(base_model_nparray):
            assert (delta.shape == base_model_nparray.shape), (
                f'Shape of delta ({delta.shape}) is not equal to shape of model'
                f' layer ({base_model_nparray.shape})'
            )
        # assert('model' in tensor_key[3]), 'The tensorkey should be provided
        # from the base model'
        # Aggregator UUID has the prefix 'aggregator'
        if 'aggregator' in origin and not creates_model:
            new_tags = change_tags(tags, remove_field='delta')
            new_model_tensor_key = TensorKey(
                tensor_name, origin, round_number, report, new_tags)
        else:
            new_model_tensor_key = TensorKey(
                tensor_name, origin, round_number, report, ('model',))

        return new_model_tensor_key, base_model_nparray + delta

    def find_dependencies(self, tensor_key, send_model_deltas):
        """Resolve the tensors required to do the specified operation."""
        tensor_key_dependencies = []

        tensor_name, origin, round_number, report, tags = tensor_key

        if 'model' in tags and send_model_deltas:
            if round_number >= 1:
                # The new model can be generated by previous model + delta
                tensor_key_dependencies.append(
                    TensorKey(
                        tensor_name, origin, round_number - 1, report, tags
                    )
                )
                if self.compression_pipeline.is_lossy():
                    new_tags = ('aggregated', 'delta', 'lossy_compressed')
                else:
                    new_tags = ('aggregated', 'delta', 'compressed')
                tensor_key_dependencies.append(
                    TensorKey(
                        tensor_name, origin, round_number, report, new_tags
                    )
                )

        return tensor_key_dependencies



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""PyTorchDataLoader module."""

from math import ceil, fabs

import numpy as np

from .loader import DataLoader


class FeTSChallengeDataLoaderWrapper(DataLoader):
    """Federation Data Loader for FeTS Challenge Model."""

    def __init__(self, data_path, feature_shape):
        self.train_dataloader = None
        self.val_dataloader = None
        self.feature_shape = feature_shape

    def set_dataloaders(self, train_dataloader, val_dataloader):
        self.train_dataloader = train_dataloader
        self.val_dataloader = val_dataloader

    def get_feature_shape(self):
        """Get the shape of an example feature array.

        Returns:
            tuple: shape of an example feature array
        """
        return self.feature_shape

    def get_train_loader(self, batch_size=None, num_batches=None):
        """
        Get training data loader.

        Returns
        -------
        loader object
        """
        return self.train_dataloader

    def get_valid_loader(self, batch_size=None):
        """
        Get validation data loader.

        Returns:
            loader object
        """
        return self.val_dataloader

    def get_train_data_size(self):
        """
        Get total number of training samples.

        Returns:
            int: number of training samples
        """
        return len(self.train_dataloader.dataset)

    def get_valid_data_size(self):
        """
        Get total number of validation samples.

        Returns:
            int: number of validation samples
        """
        return len(self.val_dataloader.dataset)



# Copyright (C) 2020-2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

"""Workspace utils module."""

import logging
import os
import shutil
import sys
import time
from pathlib import Path
from subprocess import check_call
from sys import executable
from typing import Optional
from typing import Tuple
from typing import Union

logger = logging.getLogger(__name__)


class ExperimentWorkspace:
    """Experiment workspace context manager."""

    def __init__(
            self,
            experiment_name: str,
            data_file_path: Path,
            is_install_requirements: bool = False
    ) -> None:
        """Initialize workspace context manager."""
        self.experiment_name = experiment_name
        self.data_file_path = data_file_path
        self.cwd = os.getcwd()
        self.experiment_work_dir = f'{self.cwd}/{self.experiment_name}'
        self.is_install_requirements = is_install_requirements

    def _install_requirements(self):
        """Install experiment requirements."""
        requirements_filename = f'{self.experiment_work_dir}/requirements.txt'

        if os.path.isfile(requirements_filename):
            attempts = 10
            for _ in range(attempts):
                try:
                    check_call([
                        executable, '-m', 'pip', 'install', '-r', requirements_filename],
                        shell=False)
                except Exception as exc:
                    logger.error(f'Failed to install requirements: {exc}')
                    # It's a workaround for cases when collaborators run
                    # in common virtual environment
                    time.sleep(5)
                else:
                    break
        else:
            logger.error('No ' + requirements_filename + ' file found.')

    def __enter__(self):
        """Create a collaborator workspace for the experiment."""
        if os.path.exists(self.experiment_work_dir):
            shutil.rmtree(self.experiment_work_dir)
        os.makedirs(self.experiment_work_dir)

        shutil.unpack_archive(self.data_file_path, self.experiment_work_dir, format='zip')

        if self.is_install_requirements:
            self._install_requirements()

        os.chdir(self.experiment_work_dir)

        # This is needed for python module finder
        sys.path.append(self.experiment_work_dir)

    def __exit__(self, exc_type, exc_value, traceback):
        """Remove the workspace."""
        os.chdir(self.cwd)
        shutil.rmtree(self.experiment_name, ignore_errors=True)
        if self.experiment_work_dir in sys.path:
            sys.path.remove(self.experiment_work_dir)
        os.remove(self.data_file_path)


def dump_requirements_file(
        path: Union[str, Path] = './requirements.txt',
        keep_original_prefixes: bool = True,
        prefixes: Optional[Union[Tuple[str], str]] = None,
) -> None:
    """Prepare and save requirements.txt."""
    from pip._internal.operations import freeze
    path = Path(path).absolute()

    # Prepare user provided prefixes for merge with original ones
    if prefixes is None:
        prefixes = set()
    elif type(prefixes) is str:
        prefixes = set(prefixes,)
    else:
        prefixes = set(prefixes)

    # Merge prefixes:
    # We expect that all the prefixes in a requirement file
    # are placed at the top
    if keep_original_prefixes and path.is_file():
        with open(path) as f:
            for line in f:
                if line == '\n':
                    continue
                if line[0] == '-':
                    prefixes |= {line.replace('\n', '')}
                else:
                    break

    requirements_generator = freeze.freeze()
    with open(path, 'w') as f:
        for prefix in prefixes:
            f.write(prefix + '\n')

        for package in requirements_generator:
            if _is_package_versioned(package):
                f.write(package + '\n')


def _is_package_versioned(package: str) -> bool:
    """Check if the package has a version."""
    return ('==' in package
            and package not in ['pkg-resources==0.0.0', 'pkg_resources==0.0.0']
            and '-e ' not in package
            )
